from langgraph.graph import END, START, StateGraph
from mtmaisdk.clients.rest.models import AssisantState

from mtmai.agents.joke_graph.nodes.joke_writer_node import JokeWriterNode


class JokeGraph:
    @property
    def name(self):
        return "jokeGraph"

    @property
    def description(self):
        return "笑话生成器(用于测试)"

    async def build_graph(self):
        builder = StateGraph(AssisantState)

        builder.add_node("joke_writer", JokeWriterNode())
        builder.add_edge(START, "joke_writer")
        builder.add_edge("joke_writer", END)

        return builder



from langchain.tools import tool

# from langchain_community.agent_toolkits import GmailToolkit
# from langchain_community.tools.gmail.create_draft import GmailCreateDraft


class CreateDraftTool:
    @tool("Create Draft")
    def create_draft(data):
        """
        Useful to create an email draft.
         The input to this tool should be a pipe (|) separated text
         of length 3 (three), representing who to send the email to,
         the subject of the email and the actual message.
         For example, `lorem@ipsum.com|Nice To Meet You|Hey it was great to meet you.`.
        """
        email, subject, message = data.split("|")
        # gmail = GmailToolkit()
        # draft = GmailCreateDraft(api_resource=gmail.api_resource)
        # result = draft({"to": [email], "subject": subject, "message": message})
        # return f"\nDraft created: {result}\n"
        return "fake create_draft result11111111"



from langchain_core.messages import AIMessage
from langchain_core.prompts import ChatPromptTemplate, MessagesPlaceholder
from langchain_core.runnables import Runnable, RunnableConfig, RunnableLambda

from mtmai.agents.nodes.utils import swap_roles
from mtmai.core.logging import get_logger
from mtmai.llm.llm import get_fast_llm
from mtmai.models.graph_config import InterviewState

logger = get_logger()

gen_qn_prompt = ChatPromptTemplate.from_messages(
    [
        (
            "system",
            """You are an experienced Wikipedia writer and want to edit a specific page. \
Besides your identity as a Wikipedia writer, you have a specific focus when researching the topic. \
Now, you are chatting with an expert to get information. Ask good questions to get more useful information.

When you have no more questions to ask, say "Thank you so much for your help!" to end the conversation.\
Please only ask one question at a time and don't ask what you have asked before.\
Your questions should be related to the topic you want to write.
Be comprehensive and curious, gaining as much unique insight from the expert as possible.\

Stay true to your specific perspective:

{persona}""",
        ),
        MessagesPlaceholder(variable_name="messages", optional=True),
    ]
)


def tag_with_name(ai_message: AIMessage, name: str):
    ai_message.name = name
    return ai_message


class GenQuestionNode:
    def __init__(self, runnable: Runnable):
        self.runnable = runnable

    def node_name(self):
        return "gen_question"

    async def __call__(
        self,
        state: InterviewState,
        config: RunnableConfig,
    ):
        logger.info(f"进入 gen_question node: {state}")
        editor = state["editor"]
        gn_chain = (
            RunnableLambda(swap_roles).bind(name=editor.name)
            | gen_qn_prompt.partial(persona=editor.persona)
            | get_fast_llm()
            | RunnableLambda(tag_with_name).bind(name=editor.name)
        )
        result = await gn_chain.ainvoke(state)
        return {"messages": [result]}



import json
import logging
import pprint
import time
from collections.abc import AsyncIterator, Generator

from fastapi.encoders import jsonable_encoder
from fastapi.responses import StreamingResponse
from json_repair import repair_json
from langchain_core.messages import BaseMessage
from langchain_core.output_parsers import PydanticOutputParser
from openai import Stream
from openai.types.chat.chat_completion_chunk import (
    ChatCompletionChunk,
    Choice,
    ChoiceDelta,
)
from opentelemetry import trace
from pydantic import BaseModel

from mtmai.mtlibs import mtutils

logger = logging.getLogger()
tracer = trace.get_tracer_provider().get_tracer(__name__)


def repaire_json(json_like_input: str):
    good_json_string = repair_json(json_like_input, skip_json_loads=True)
    return good_json_string


def chat_complations_stream_text(response: Stream[ChatCompletionChunk]):
    """
    兼容 vercel ai sdk
    等同于 nextjs /api/chat/route.ts 中的:
        return await streamText({
            model: getAiModalDefault(),
            messages,
        }).toDataStreamResponse();
    """
    for chunk in response:
        if not chunk.choices[0].finish_reason:
            if chunk.choices[0].delta.content:
                yield f'{chunk.choices[0].index}: "{chunk.choices[0].delta.content}"\n'
                # yield f"data: {json.dumps(chunk2)}\n"
        else:
            # 结束
            final_chunk = {
                "id": chunk.id,
                "object": chunk.object,
                "created": chunk.created,
                "model": chunk.model,
                "finishReason": chunk.choices[0].finish_reason,
                "usage": jsonable_encoder(chunk.usage),
            }
            yield f"d: {json.dumps(final_chunk)}\n"
            # 明确的结束符
            yield "[DONE]\n"


def stream_response(stream_chunck: Stream[ChatCompletionChunk]):
    def gen_stream():
        for chunk in stream_chunck:
            pprint.pp(chunk)
            yield f"data: {json.dumps(jsonable_encoder( chunk))}\n\n"
            if chunk.choices[0].finish_reason is not None:
                yield "data: [DONE]\n"

    return StreamingResponse(gen_stream(), media_type="text/event-stream")


async def stream_text(stream: AsyncIterator[BaseMessage]):
    async for ai_message_chunk in stream:
        if ai_message_chunk.content:
            yield f"0:{json.dumps(ai_message_chunk.content)} \n"


def gen_text_stream(words: Generator[str, None]):
    """以stream 的方式向chat 输出字符串 (旧代码)"""
    for w in words:
        chat_chunk = ChatCompletionChunk(
            id=mtutils.gen_orm_id_key(),
            object="chat.completion.chunk",
            created=int(time.time()),
            model="agent",
            choices=[
                Choice(
                    index=0,
                    delta=ChoiceDelta(content=w, role="assistant", text=w),
                )
            ],
        )
        yield f"data: {json.dumps(jsonable_encoder(chat_chunk))}\n\n"


class ClientAttachment(BaseModel):
    name: str
    contentType: str
    url: str


class ToolInvocation(BaseModel):
    toolCallId: str
    toolName: str
    args: dict
    result: dict


class ClientMessage(BaseModel):
    role: str
    content: str
    experimental_attachments: list[ClientAttachment] | None = None
    toolInvocations: list[ToolInvocation] | None = None


class ClientAttachment(BaseModel):
    name: str
    contentType: str
    url: str


class ToolInvocation(BaseModel):
    toolCallId: str
    toolName: str
    args: dict
    result: dict


def gen_text_stream_2(words: Generator[str, None]):
    """(可能是旧代码)以stream 的方式向chat 输出字符串"""
    for w in words:
        chat_chunk = ChatCompletionChunk(
            id=mtutils.gen_orm_id_key(),
            object="chat.completion.chunk",
            created=int(time.time()),
            model="agent",
            choices=[
                Choice(
                    index=0,
                    delta=ChoiceDelta(content=w, role="assistant", text=w),
                )
            ],
        )
        yield f"data: {json.dumps(jsonable_encoder(chat_chunk))}\n\n"

    # 发送结束标志
    end_chunk = ChatCompletionChunk(
        id=mtutils.gen_orm_id_key(),
        object="chat.completion.chunk",
        created=int(time.time()),
        model="agent",
        choices=[
            Choice(
                index=0,
                delta=ChoiceDelta(content="", role="assistant"),
                finish_reason="stop",
            )
        ],
    )
    yield f"data: {json.dumps(jsonable_encoder(end_chunk))}\n\n"


def convert_to_openai_messages(messages: list[ClientMessage]):
    openai_messages = []

    for message in messages:
        parts = []

        parts.append({"type": "text", "text": message.content})

        if message.experimental_attachments:
            for attachment in message.experimental_attachments:
                if attachment.contentType.startswith("image"):
                    parts.append(
                        {"type": "image_url", "image_url": {"url": attachment.url}}
                    )

                elif attachment.contentType.startswith("text"):
                    parts.append({"type": "text", "text": attachment.url})

        if message.toolInvocations:
            tool_calls = [
                {
                    "id": tool_invocation.toolCallId,
                    "type": "function",
                    "function": {
                        "name": tool_invocation.toolName,
                        "arguments": json.dumps(tool_invocation.args),
                    },
                }
                for tool_invocation in message.toolInvocations
            ]

            openai_messages.append({"role": "assistant", "tool_calls": tool_calls})

            tool_results = [
                {
                    "role": "tool",
                    "content": json.dumps(tool_invocation.result),
                    "tool_call_id": tool_invocation.toolCallId,
                }
                for tool_invocation in message.toolInvocations
            ]

            openai_messages.extend(tool_results)

            continue

        openai_messages.append({"role": message.role, "content": parts})

    return openai_messages


def get_json_format_instructions(pydantic_model: BaseModel):
    """
    从pydantic 模型中获取 json 格式的格式化说明,(使用了增强提示词)
    """
    parser = PydanticOutputParser(pydantic_object=pydantic_model)
    format_instructions = parser.get_format_instructions()
    format_instructions = (
        format_instructions
        + "\n\nDouble-check your output to ensure it is valid JSON before submitting.\n"
    )
    return format_instructions



import torch
import torch.nn as nn
import torch.nn.functional as F
import numpy as np



# 超参数设置
class Config(object):

    """配置参数"""
    def __init__(self, dataset, embedding):
        self.model_name = 'TextRNN'
        self.train_path = dataset + '/data/train.txt'                                # 训练集
        self.dev_path = dataset + '/data/dev.txt'                                    # 验证集
        self.test_path = dataset + '/data/test.txt'                                  # 测试集
        self.class_list = [x.strip() for x in open(
            dataset + '/data/class.txt').readlines()]                                # 类别列表
        self.vocab_path = dataset + '/data/vocab.pkl'                                # 词表
        self.save_path = dataset + '/saved_dict/' + self.model_name + '.ckpt'        # 模型训练结果
        self.embedding_pretrained = torch.tensor(
            np.load(dataset + '/data/' + embedding)["embeddings"].astype('float32')) # 预训练词向量
        self.device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')   # 设备

        self.dropout = 0.5                                              # 随机失活
        self.require_improvement = 1000                                 # 若超过1000batch效果还没提升，则提前结束训练
        self.num_classes = len(self.class_list)                         # 类别数
        self.n_vocab = 0                                                # 词表大小，在运行时赋值
        self.num_epochs = 10                                            # epoch数
        self.batch_size = 128                                           # mini-batch大小
        self.pad_size = 32                                              # 每句话处理成的长度(短填长切)
        self.learning_rate = 1e-3                                       # 学习率
        self.embed = self.embedding_pretrained.size(1)                  # 词向量维度, 若使用了预训练词向量，则维度统一
        self.hidden_size = 128                                          # lstm隐藏层
        self.num_layers = 2                                             # lstm层数


'''Recurrent Neural Network for Text Classification with Multi-Task Learning'''


class RNNModel(nn.Module):
    def __init__(self, config):
        super(RNNModel, self).__init__()
        # 使用预训练的词向量模型，freeze=False 表示允许参数在训练中更新
        self.embedding = nn.Embedding.from_pretrained(config.embedding_pretrained, freeze=False)
        # bidirectional=True表示使用的是双向LSTM
        self.lstm = nn.LSTM(config.embed, config.hidden_size, config.num_layers,
                            bidirectional=True, batch_first=True, dropout=config.dropout)
        # 因为是双向LSTM，所以层数为config.hidden_size * 2
        self.fc = nn.Linear(config.hidden_size * 2, config.num_classes)

    def forward(self, x):
        out = self.embedding(x)  # [batch_size, seq_len, embeding]=[128, 32, 300]
        # lstm 的input为[batchsize, max_length, embedding_size]，输出表示为 output,(h_n,c_n),
        # 保存了每个时间步的输出，如果想要获取最后一个时间步的输出，则可以这么获取：output_last = output[:,-1,:]
        out, _ = self.lstm(out)
        out = self.fc(out[:, -1, :])  # 句子最后时刻的 hidden state
        return out



# coding: utf-8

"""
    Mtmai API

    The Mtmai API

    The version of the OpenAPI document: 1.0.0
    Generated by OpenAPI Generator (https://openapi-generator.tech)

    Do not edit the class manually.
"""  # noqa: E501

import warnings
from pydantic import validate_call, Field, StrictFloat, StrictStr, StrictInt
from typing import Any, Dict, List, Optional, Tuple, Union
from typing_extensions import Annotated

from pydantic import Field
from typing import Any, Dict, Optional
from typing_extensions import Annotated
from mtmai.gomtmclients.rest.models.event_update_cancel200_response import EventUpdateCancel200Response
from mtmai.gomtmclients.rest.models.replay_workflow_runs_request import ReplayWorkflowRunsRequest
from mtmai.gomtmclients.rest.models.replay_workflow_runs_response import ReplayWorkflowRunsResponse
from mtmai.gomtmclients.rest.models.trigger_workflow_run_request import TriggerWorkflowRunRequest
from mtmai.gomtmclients.rest.models.workflow_run import WorkflowRun
from mtmai.gomtmclients.rest.models.workflow_runs_cancel_request import WorkflowRunsCancelRequest

from mtmai.gomtmclients.rest.api_client import ApiClient, RequestSerialized
from mtmai.gomtmclients.rest.api_response import ApiResponse
from mtmai.gomtmclients.rest.rest import RESTResponseType


class WorkflowRunApi:
    """NOTE: This class is auto generated by OpenAPI Generator
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    def __init__(self, api_client=None) -> None:
        if api_client is None:
            api_client = ApiClient.get_default()
        self.api_client = api_client


    @validate_call
    async def workflow_run_cancel(
        self,
        tenant: Annotated[str, Field(min_length=36, strict=True, max_length=36, description="The tenant id")],
        workflow_runs_cancel_request: Annotated[WorkflowRunsCancelRequest, Field(description="The input to cancel the workflow runs")],
        _request_timeout: Union[
            None,
            Annotated[StrictFloat, Field(gt=0)],
            Tuple[
                Annotated[StrictFloat, Field(gt=0)],
                Annotated[StrictFloat, Field(gt=0)]
            ]
        ] = None,
        _request_auth: Optional[Dict[StrictStr, Any]] = None,
        _content_type: Optional[StrictStr] = None,
        _headers: Optional[Dict[StrictStr, Any]] = None,
        _host_index: Annotated[StrictInt, Field(ge=0, le=0)] = 0,
    ) -> EventUpdateCancel200Response:
        """Cancel workflow runs

        Cancel a batch of workflow runs

        :param tenant: The tenant id (required)
        :type tenant: str
        :param workflow_runs_cancel_request: The input to cancel the workflow runs (required)
        :type workflow_runs_cancel_request: WorkflowRunsCancelRequest
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :type _request_timeout: int, tuple(int, int), optional
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the
                              authentication in the spec for a single request.
        :type _request_auth: dict, optional
        :param _content_type: force content-type for the request.
        :type _content_type: str, Optional
        :param _headers: set to override the headers for a single
                         request; this effectively ignores the headers
                         in the spec for a single request.
        :type _headers: dict, optional
        :param _host_index: set to override the host_index for a single
                            request; this effectively ignores the host_index
                            in the spec for a single request.
        :type _host_index: int, optional
        :return: Returns the result object.
        """ # noqa: E501

        _param = self._workflow_run_cancel_serialize(
            tenant=tenant,
            workflow_runs_cancel_request=workflow_runs_cancel_request,
            _request_auth=_request_auth,
            _content_type=_content_type,
            _headers=_headers,
            _host_index=_host_index
        )

        _response_types_map: Dict[str, Optional[str]] = {
            '200': "EventUpdateCancel200Response",
            '400': "APIErrors",
            '403': "APIErrors",
        }
        response_data = await self.api_client.call_api(
            *_param,
            _request_timeout=_request_timeout
        )
        await response_data.read()
        return self.api_client.response_deserialize(
            response_data=response_data,
            response_types_map=_response_types_map,
        ).data


    @validate_call
    async def workflow_run_cancel_with_http_info(
        self,
        tenant: Annotated[str, Field(min_length=36, strict=True, max_length=36, description="The tenant id")],
        workflow_runs_cancel_request: Annotated[WorkflowRunsCancelRequest, Field(description="The input to cancel the workflow runs")],
        _request_timeout: Union[
            None,
            Annotated[StrictFloat, Field(gt=0)],
            Tuple[
                Annotated[StrictFloat, Field(gt=0)],
                Annotated[StrictFloat, Field(gt=0)]
            ]
        ] = None,
        _request_auth: Optional[Dict[StrictStr, Any]] = None,
        _content_type: Optional[StrictStr] = None,
        _headers: Optional[Dict[StrictStr, Any]] = None,
        _host_index: Annotated[StrictInt, Field(ge=0, le=0)] = 0,
    ) -> ApiResponse[EventUpdateCancel200Response]:
        """Cancel workflow runs

        Cancel a batch of workflow runs

        :param tenant: The tenant id (required)
        :type tenant: str
        :param workflow_runs_cancel_request: The input to cancel the workflow runs (required)
        :type workflow_runs_cancel_request: WorkflowRunsCancelRequest
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :type _request_timeout: int, tuple(int, int), optional
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the
                              authentication in the spec for a single request.
        :type _request_auth: dict, optional
        :param _content_type: force content-type for the request.
        :type _content_type: str, Optional
        :param _headers: set to override the headers for a single
                         request; this effectively ignores the headers
                         in the spec for a single request.
        :type _headers: dict, optional
        :param _host_index: set to override the host_index for a single
                            request; this effectively ignores the host_index
                            in the spec for a single request.
        :type _host_index: int, optional
        :return: Returns the result object.
        """ # noqa: E501

        _param = self._workflow_run_cancel_serialize(
            tenant=tenant,
            workflow_runs_cancel_request=workflow_runs_cancel_request,
            _request_auth=_request_auth,
            _content_type=_content_type,
            _headers=_headers,
            _host_index=_host_index
        )

        _response_types_map: Dict[str, Optional[str]] = {
            '200': "EventUpdateCancel200Response",
            '400': "APIErrors",
            '403': "APIErrors",
        }
        response_data = await self.api_client.call_api(
            *_param,
            _request_timeout=_request_timeout
        )
        await response_data.read()
        return self.api_client.response_deserialize(
            response_data=response_data,
            response_types_map=_response_types_map,
        )


    @validate_call
    async def workflow_run_cancel_without_preload_content(
        self,
        tenant: Annotated[str, Field(min_length=36, strict=True, max_length=36, description="The tenant id")],
        workflow_runs_cancel_request: Annotated[WorkflowRunsCancelRequest, Field(description="The input to cancel the workflow runs")],
        _request_timeout: Union[
            None,
            Annotated[StrictFloat, Field(gt=0)],
            Tuple[
                Annotated[StrictFloat, Field(gt=0)],
                Annotated[StrictFloat, Field(gt=0)]
            ]
        ] = None,
        _request_auth: Optional[Dict[StrictStr, Any]] = None,
        _content_type: Optional[StrictStr] = None,
        _headers: Optional[Dict[StrictStr, Any]] = None,
        _host_index: Annotated[StrictInt, Field(ge=0, le=0)] = 0,
    ) -> RESTResponseType:
        """Cancel workflow runs

        Cancel a batch of workflow runs

        :param tenant: The tenant id (required)
        :type tenant: str
        :param workflow_runs_cancel_request: The input to cancel the workflow runs (required)
        :type workflow_runs_cancel_request: WorkflowRunsCancelRequest
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :type _request_timeout: int, tuple(int, int), optional
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the
                              authentication in the spec for a single request.
        :type _request_auth: dict, optional
        :param _content_type: force content-type for the request.
        :type _content_type: str, Optional
        :param _headers: set to override the headers for a single
                         request; this effectively ignores the headers
                         in the spec for a single request.
        :type _headers: dict, optional
        :param _host_index: set to override the host_index for a single
                            request; this effectively ignores the host_index
                            in the spec for a single request.
        :type _host_index: int, optional
        :return: Returns the result object.
        """ # noqa: E501

        _param = self._workflow_run_cancel_serialize(
            tenant=tenant,
            workflow_runs_cancel_request=workflow_runs_cancel_request,
            _request_auth=_request_auth,
            _content_type=_content_type,
            _headers=_headers,
            _host_index=_host_index
        )

        _response_types_map: Dict[str, Optional[str]] = {
            '200': "EventUpdateCancel200Response",
            '400': "APIErrors",
            '403': "APIErrors",
        }
        response_data = await self.api_client.call_api(
            *_param,
            _request_timeout=_request_timeout
        )
        return response_data.response


    def _workflow_run_cancel_serialize(
        self,
        tenant,
        workflow_runs_cancel_request,
        _request_auth,
        _content_type,
        _headers,
        _host_index,
    ) -> RequestSerialized:

        _host = None

        _collection_formats: Dict[str, str] = {
        }

        _path_params: Dict[str, str] = {}
        _query_params: List[Tuple[str, str]] = []
        _header_params: Dict[str, Optional[str]] = _headers or {}
        _form_params: List[Tuple[str, str]] = []
        _files: Dict[
            str, Union[str, bytes, List[str], List[bytes], List[Tuple[str, bytes]]]
        ] = {}
        _body_params: Optional[bytes] = None

        # process the path parameters
        if tenant is not None:
            _path_params['tenant'] = tenant
        # process the query parameters
        # process the header parameters
        # process the form parameters
        # process the body parameter
        if workflow_runs_cancel_request is not None:
            _body_params = workflow_runs_cancel_request


        # set the HTTP header `Accept`
        if 'Accept' not in _header_params:
            _header_params['Accept'] = self.api_client.select_header_accept(
                [
                    'application/json'
                ]
            )

        # set the HTTP header `Content-Type`
        if _content_type:
            _header_params['Content-Type'] = _content_type
        else:
            _default_content_type = (
                self.api_client.select_header_content_type(
                    [
                        'application/json'
                    ]
                )
            )
            if _default_content_type is not None:
                _header_params['Content-Type'] = _default_content_type

        # authentication setting
        _auth_settings: List[str] = [
            'basicAuth', 
            'cookieAuth', 
            'bearerAuth'
        ]

        return self.api_client.param_serialize(
            method='POST',
            resource_path='/api/v1/tenants/{tenant}/workflows/cancel',
            path_params=_path_params,
            query_params=_query_params,
            header_params=_header_params,
            body=_body_params,
            post_params=_form_params,
            files=_files,
            auth_settings=_auth_settings,
            collection_formats=_collection_formats,
            _host=_host,
            _request_auth=_request_auth
        )




    @validate_call
    async def workflow_run_create(
        self,
        workflow: Annotated[str, Field(min_length=36, strict=True, max_length=36, description="The workflow id")],
        trigger_workflow_run_request: Annotated[TriggerWorkflowRunRequest, Field(description="The input to the workflow run")],
        version: Annotated[Optional[Annotated[str, Field(min_length=36, strict=True, max_length=36)]], Field(description="The workflow version. If not supplied, the latest version is fetched.")] = None,
        _request_timeout: Union[
            None,
            Annotated[StrictFloat, Field(gt=0)],
            Tuple[
                Annotated[StrictFloat, Field(gt=0)],
                Annotated[StrictFloat, Field(gt=0)]
            ]
        ] = None,
        _request_auth: Optional[Dict[StrictStr, Any]] = None,
        _content_type: Optional[StrictStr] = None,
        _headers: Optional[Dict[StrictStr, Any]] = None,
        _host_index: Annotated[StrictInt, Field(ge=0, le=0)] = 0,
    ) -> WorkflowRun:
        """Trigger workflow run

        Trigger a new workflow run for a tenant

        :param workflow: The workflow id (required)
        :type workflow: str
        :param trigger_workflow_run_request: The input to the workflow run (required)
        :type trigger_workflow_run_request: TriggerWorkflowRunRequest
        :param version: The workflow version. If not supplied, the latest version is fetched.
        :type version: str
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :type _request_timeout: int, tuple(int, int), optional
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the
                              authentication in the spec for a single request.
        :type _request_auth: dict, optional
        :param _content_type: force content-type for the request.
        :type _content_type: str, Optional
        :param _headers: set to override the headers for a single
                         request; this effectively ignores the headers
                         in the spec for a single request.
        :type _headers: dict, optional
        :param _host_index: set to override the host_index for a single
                            request; this effectively ignores the host_index
                            in the spec for a single request.
        :type _host_index: int, optional
        :return: Returns the result object.
        """ # noqa: E501

        _param = self._workflow_run_create_serialize(
            workflow=workflow,
            trigger_workflow_run_request=trigger_workflow_run_request,
            version=version,
            _request_auth=_request_auth,
            _content_type=_content_type,
            _headers=_headers,
            _host_index=_host_index
        )

        _response_types_map: Dict[str, Optional[str]] = {
            '200': "WorkflowRun",
            '400': "APIErrors",
            '403': "APIErrors",
            '404': "APIErrors",
            '429': "APIErrors",
        }
        response_data = await self.api_client.call_api(
            *_param,
            _request_timeout=_request_timeout
        )
        await response_data.read()
        return self.api_client.response_deserialize(
            response_data=response_data,
            response_types_map=_response_types_map,
        ).data


    @validate_call
    async def workflow_run_create_with_http_info(
        self,
        workflow: Annotated[str, Field(min_length=36, strict=True, max_length=36, description="The workflow id")],
        trigger_workflow_run_request: Annotated[TriggerWorkflowRunRequest, Field(description="The input to the workflow run")],
        version: Annotated[Optional[Annotated[str, Field(min_length=36, strict=True, max_length=36)]], Field(description="The workflow version. If not supplied, the latest version is fetched.")] = None,
        _request_timeout: Union[
            None,
            Annotated[StrictFloat, Field(gt=0)],
            Tuple[
                Annotated[StrictFloat, Field(gt=0)],
                Annotated[StrictFloat, Field(gt=0)]
            ]
        ] = None,
        _request_auth: Optional[Dict[StrictStr, Any]] = None,
        _content_type: Optional[StrictStr] = None,
        _headers: Optional[Dict[StrictStr, Any]] = None,
        _host_index: Annotated[StrictInt, Field(ge=0, le=0)] = 0,
    ) -> ApiResponse[WorkflowRun]:
        """Trigger workflow run

        Trigger a new workflow run for a tenant

        :param workflow: The workflow id (required)
        :type workflow: str
        :param trigger_workflow_run_request: The input to the workflow run (required)
        :type trigger_workflow_run_request: TriggerWorkflowRunRequest
        :param version: The workflow version. If not supplied, the latest version is fetched.
        :type version: str
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :type _request_timeout: int, tuple(int, int), optional
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the
                              authentication in the spec for a single request.
        :type _request_auth: dict, optional
        :param _content_type: force content-type for the request.
        :type _content_type: str, Optional
        :param _headers: set to override the headers for a single
                         request; this effectively ignores the headers
                         in the spec for a single request.
        :type _headers: dict, optional
        :param _host_index: set to override the host_index for a single
                            request; this effectively ignores the host_index
                            in the spec for a single request.
        :type _host_index: int, optional
        :return: Returns the result object.
        """ # noqa: E501

        _param = self._workflow_run_create_serialize(
            workflow=workflow,
            trigger_workflow_run_request=trigger_workflow_run_request,
            version=version,
            _request_auth=_request_auth,
            _content_type=_content_type,
            _headers=_headers,
            _host_index=_host_index
        )

        _response_types_map: Dict[str, Optional[str]] = {
            '200': "WorkflowRun",
            '400': "APIErrors",
            '403': "APIErrors",
            '404': "APIErrors",
            '429': "APIErrors",
        }
        response_data = await self.api_client.call_api(
            *_param,
            _request_timeout=_request_timeout
        )
        await response_data.read()
        return self.api_client.response_deserialize(
            response_data=response_data,
            response_types_map=_response_types_map,
        )


    @validate_call
    async def workflow_run_create_without_preload_content(
        self,
        workflow: Annotated[str, Field(min_length=36, strict=True, max_length=36, description="The workflow id")],
        trigger_workflow_run_request: Annotated[TriggerWorkflowRunRequest, Field(description="The input to the workflow run")],
        version: Annotated[Optional[Annotated[str, Field(min_length=36, strict=True, max_length=36)]], Field(description="The workflow version. If not supplied, the latest version is fetched.")] = None,
        _request_timeout: Union[
            None,
            Annotated[StrictFloat, Field(gt=0)],
            Tuple[
                Annotated[StrictFloat, Field(gt=0)],
                Annotated[StrictFloat, Field(gt=0)]
            ]
        ] = None,
        _request_auth: Optional[Dict[StrictStr, Any]] = None,
        _content_type: Optional[StrictStr] = None,
        _headers: Optional[Dict[StrictStr, Any]] = None,
        _host_index: Annotated[StrictInt, Field(ge=0, le=0)] = 0,
    ) -> RESTResponseType:
        """Trigger workflow run

        Trigger a new workflow run for a tenant

        :param workflow: The workflow id (required)
        :type workflow: str
        :param trigger_workflow_run_request: The input to the workflow run (required)
        :type trigger_workflow_run_request: TriggerWorkflowRunRequest
        :param version: The workflow version. If not supplied, the latest version is fetched.
        :type version: str
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :type _request_timeout: int, tuple(int, int), optional
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the
                              authentication in the spec for a single request.
        :type _request_auth: dict, optional
        :param _content_type: force content-type for the request.
        :type _content_type: str, Optional
        :param _headers: set to override the headers for a single
                         request; this effectively ignores the headers
                         in the spec for a single request.
        :type _headers: dict, optional
        :param _host_index: set to override the host_index for a single
                            request; this effectively ignores the host_index
                            in the spec for a single request.
        :type _host_index: int, optional
        :return: Returns the result object.
        """ # noqa: E501

        _param = self._workflow_run_create_serialize(
            workflow=workflow,
            trigger_workflow_run_request=trigger_workflow_run_request,
            version=version,
            _request_auth=_request_auth,
            _content_type=_content_type,
            _headers=_headers,
            _host_index=_host_index
        )

        _response_types_map: Dict[str, Optional[str]] = {
            '200': "WorkflowRun",
            '400': "APIErrors",
            '403': "APIErrors",
            '404': "APIErrors",
            '429': "APIErrors",
        }
        response_data = await self.api_client.call_api(
            *_param,
            _request_timeout=_request_timeout
        )
        return response_data.response


    def _workflow_run_create_serialize(
        self,
        workflow,
        trigger_workflow_run_request,
        version,
        _request_auth,
        _content_type,
        _headers,
        _host_index,
    ) -> RequestSerialized:

        _host = None

        _collection_formats: Dict[str, str] = {
        }

        _path_params: Dict[str, str] = {}
        _query_params: List[Tuple[str, str]] = []
        _header_params: Dict[str, Optional[str]] = _headers or {}
        _form_params: List[Tuple[str, str]] = []
        _files: Dict[
            str, Union[str, bytes, List[str], List[bytes], List[Tuple[str, bytes]]]
        ] = {}
        _body_params: Optional[bytes] = None

        # process the path parameters
        if workflow is not None:
            _path_params['workflow'] = workflow
        # process the query parameters
        if version is not None:
            
            _query_params.append(('version', version))
            
        # process the header parameters
        # process the form parameters
        # process the body parameter
        if trigger_workflow_run_request is not None:
            _body_params = trigger_workflow_run_request


        # set the HTTP header `Accept`
        if 'Accept' not in _header_params:
            _header_params['Accept'] = self.api_client.select_header_accept(
                [
                    'application/json'
                ]
            )

        # set the HTTP header `Content-Type`
        if _content_type:
            _header_params['Content-Type'] = _content_type
        else:
            _default_content_type = (
                self.api_client.select_header_content_type(
                    [
                        'application/json'
                    ]
                )
            )
            if _default_content_type is not None:
                _header_params['Content-Type'] = _default_content_type

        # authentication setting
        _auth_settings: List[str] = [
            'basicAuth', 
            'cookieAuth', 
            'bearerAuth'
        ]

        return self.api_client.param_serialize(
            method='POST',
            resource_path='/api/v1/workflows/{workflow}/trigger',
            path_params=_path_params,
            query_params=_query_params,
            header_params=_header_params,
            body=_body_params,
            post_params=_form_params,
            files=_files,
            auth_settings=_auth_settings,
            collection_formats=_collection_formats,
            _host=_host,
            _request_auth=_request_auth
        )




    @validate_call
    async def workflow_run_get_input(
        self,
        tenant: Annotated[str, Field(min_length=36, strict=True, max_length=36, description="The tenant id")],
        workflow_run: Annotated[str, Field(min_length=36, strict=True, max_length=36, description="The workflow run id")],
        _request_timeout: Union[
            None,
            Annotated[StrictFloat, Field(gt=0)],
            Tuple[
                Annotated[StrictFloat, Field(gt=0)],
                Annotated[StrictFloat, Field(gt=0)]
            ]
        ] = None,
        _request_auth: Optional[Dict[StrictStr, Any]] = None,
        _content_type: Optional[StrictStr] = None,
        _headers: Optional[Dict[StrictStr, Any]] = None,
        _host_index: Annotated[StrictInt, Field(ge=0, le=0)] = 0,
    ) -> Dict[str, object]:
        """Get workflow run input

        Get the input for a workflow run.

        :param tenant: The tenant id (required)
        :type tenant: str
        :param workflow_run: The workflow run id (required)
        :type workflow_run: str
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :type _request_timeout: int, tuple(int, int), optional
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the
                              authentication in the spec for a single request.
        :type _request_auth: dict, optional
        :param _content_type: force content-type for the request.
        :type _content_type: str, Optional
        :param _headers: set to override the headers for a single
                         request; this effectively ignores the headers
                         in the spec for a single request.
        :type _headers: dict, optional
        :param _host_index: set to override the host_index for a single
                            request; this effectively ignores the host_index
                            in the spec for a single request.
        :type _host_index: int, optional
        :return: Returns the result object.
        """ # noqa: E501

        _param = self._workflow_run_get_input_serialize(
            tenant=tenant,
            workflow_run=workflow_run,
            _request_auth=_request_auth,
            _content_type=_content_type,
            _headers=_headers,
            _host_index=_host_index
        )

        _response_types_map: Dict[str, Optional[str]] = {
            '200': "Dict[str, object]",
            '400': "APIErrors",
            '403': "APIErrors",
            '404': "APIErrors",
        }
        response_data = await self.api_client.call_api(
            *_param,
            _request_timeout=_request_timeout
        )
        await response_data.read()
        return self.api_client.response_deserialize(
            response_data=response_data,
            response_types_map=_response_types_map,
        ).data


    @validate_call
    async def workflow_run_get_input_with_http_info(
        self,
        tenant: Annotated[str, Field(min_length=36, strict=True, max_length=36, description="The tenant id")],
        workflow_run: Annotated[str, Field(min_length=36, strict=True, max_length=36, description="The workflow run id")],
        _request_timeout: Union[
            None,
            Annotated[StrictFloat, Field(gt=0)],
            Tuple[
                Annotated[StrictFloat, Field(gt=0)],
                Annotated[StrictFloat, Field(gt=0)]
            ]
        ] = None,
        _request_auth: Optional[Dict[StrictStr, Any]] = None,
        _content_type: Optional[StrictStr] = None,
        _headers: Optional[Dict[StrictStr, Any]] = None,
        _host_index: Annotated[StrictInt, Field(ge=0, le=0)] = 0,
    ) -> ApiResponse[Dict[str, object]]:
        """Get workflow run input

        Get the input for a workflow run.

        :param tenant: The tenant id (required)
        :type tenant: str
        :param workflow_run: The workflow run id (required)
        :type workflow_run: str
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :type _request_timeout: int, tuple(int, int), optional
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the
                              authentication in the spec for a single request.
        :type _request_auth: dict, optional
        :param _content_type: force content-type for the request.
        :type _content_type: str, Optional
        :param _headers: set to override the headers for a single
                         request; this effectively ignores the headers
                         in the spec for a single request.
        :type _headers: dict, optional
        :param _host_index: set to override the host_index for a single
                            request; this effectively ignores the host_index
                            in the spec for a single request.
        :type _host_index: int, optional
        :return: Returns the result object.
        """ # noqa: E501

        _param = self._workflow_run_get_input_serialize(
            tenant=tenant,
            workflow_run=workflow_run,
            _request_auth=_request_auth,
            _content_type=_content_type,
            _headers=_headers,
            _host_index=_host_index
        )

        _response_types_map: Dict[str, Optional[str]] = {
            '200': "Dict[str, object]",
            '400': "APIErrors",
            '403': "APIErrors",
            '404': "APIErrors",
        }
        response_data = await self.api_client.call_api(
            *_param,
            _request_timeout=_request_timeout
        )
        await response_data.read()
        return self.api_client.response_deserialize(
            response_data=response_data,
            response_types_map=_response_types_map,
        )


    @validate_call
    async def workflow_run_get_input_without_preload_content(
        self,
        tenant: Annotated[str, Field(min_length=36, strict=True, max_length=36, description="The tenant id")],
        workflow_run: Annotated[str, Field(min_length=36, strict=True, max_length=36, description="The workflow run id")],
        _request_timeout: Union[
            None,
            Annotated[StrictFloat, Field(gt=0)],
            Tuple[
                Annotated[StrictFloat, Field(gt=0)],
                Annotated[StrictFloat, Field(gt=0)]
            ]
        ] = None,
        _request_auth: Optional[Dict[StrictStr, Any]] = None,
        _content_type: Optional[StrictStr] = None,
        _headers: Optional[Dict[StrictStr, Any]] = None,
        _host_index: Annotated[StrictInt, Field(ge=0, le=0)] = 0,
    ) -> RESTResponseType:
        """Get workflow run input

        Get the input for a workflow run.

        :param tenant: The tenant id (required)
        :type tenant: str
        :param workflow_run: The workflow run id (required)
        :type workflow_run: str
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :type _request_timeout: int, tuple(int, int), optional
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the
                              authentication in the spec for a single request.
        :type _request_auth: dict, optional
        :param _content_type: force content-type for the request.
        :type _content_type: str, Optional
        :param _headers: set to override the headers for a single
                         request; this effectively ignores the headers
                         in the spec for a single request.
        :type _headers: dict, optional
        :param _host_index: set to override the host_index for a single
                            request; this effectively ignores the host_index
                            in the spec for a single request.
        :type _host_index: int, optional
        :return: Returns the result object.
        """ # noqa: E501

        _param = self._workflow_run_get_input_serialize(
            tenant=tenant,
            workflow_run=workflow_run,
            _request_auth=_request_auth,
            _content_type=_content_type,
            _headers=_headers,
            _host_index=_host_index
        )

        _response_types_map: Dict[str, Optional[str]] = {
            '200': "Dict[str, object]",
            '400': "APIErrors",
            '403': "APIErrors",
            '404': "APIErrors",
        }
        response_data = await self.api_client.call_api(
            *_param,
            _request_timeout=_request_timeout
        )
        return response_data.response


    def _workflow_run_get_input_serialize(
        self,
        tenant,
        workflow_run,
        _request_auth,
        _content_type,
        _headers,
        _host_index,
    ) -> RequestSerialized:

        _host = None

        _collection_formats: Dict[str, str] = {
        }

        _path_params: Dict[str, str] = {}
        _query_params: List[Tuple[str, str]] = []
        _header_params: Dict[str, Optional[str]] = _headers or {}
        _form_params: List[Tuple[str, str]] = []
        _files: Dict[
            str, Union[str, bytes, List[str], List[bytes], List[Tuple[str, bytes]]]
        ] = {}
        _body_params: Optional[bytes] = None

        # process the path parameters
        if tenant is not None:
            _path_params['tenant'] = tenant
        if workflow_run is not None:
            _path_params['workflow-run'] = workflow_run
        # process the query parameters
        # process the header parameters
        # process the form parameters
        # process the body parameter


        # set the HTTP header `Accept`
        if 'Accept' not in _header_params:
            _header_params['Accept'] = self.api_client.select_header_accept(
                [
                    'application/json'
                ]
            )


        # authentication setting
        _auth_settings: List[str] = [
            'basicAuth', 
            'cookieAuth', 
            'bearerAuth'
        ]

        return self.api_client.param_serialize(
            method='GET',
            resource_path='/api/v1/tenants/{tenant}/workflow-runs/{workflow-run}/input',
            path_params=_path_params,
            query_params=_query_params,
            header_params=_header_params,
            body=_body_params,
            post_params=_form_params,
            files=_files,
            auth_settings=_auth_settings,
            collection_formats=_collection_formats,
            _host=_host,
            _request_auth=_request_auth
        )




    @validate_call
    async def workflow_run_update_replay(
        self,
        tenant: Annotated[str, Field(min_length=36, strict=True, max_length=36, description="The tenant id")],
        replay_workflow_runs_request: Annotated[ReplayWorkflowRunsRequest, Field(description="The workflow run ids to replay")],
        _request_timeout: Union[
            None,
            Annotated[StrictFloat, Field(gt=0)],
            Tuple[
                Annotated[StrictFloat, Field(gt=0)],
                Annotated[StrictFloat, Field(gt=0)]
            ]
        ] = None,
        _request_auth: Optional[Dict[StrictStr, Any]] = None,
        _content_type: Optional[StrictStr] = None,
        _headers: Optional[Dict[StrictStr, Any]] = None,
        _host_index: Annotated[StrictInt, Field(ge=0, le=0)] = 0,
    ) -> ReplayWorkflowRunsResponse:
        """Replay workflow runs

        Replays a list of workflow runs.

        :param tenant: The tenant id (required)
        :type tenant: str
        :param replay_workflow_runs_request: The workflow run ids to replay (required)
        :type replay_workflow_runs_request: ReplayWorkflowRunsRequest
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :type _request_timeout: int, tuple(int, int), optional
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the
                              authentication in the spec for a single request.
        :type _request_auth: dict, optional
        :param _content_type: force content-type for the request.
        :type _content_type: str, Optional
        :param _headers: set to override the headers for a single
                         request; this effectively ignores the headers
                         in the spec for a single request.
        :type _headers: dict, optional
        :param _host_index: set to override the host_index for a single
                            request; this effectively ignores the host_index
                            in the spec for a single request.
        :type _host_index: int, optional
        :return: Returns the result object.
        """ # noqa: E501

        _param = self._workflow_run_update_replay_serialize(
            tenant=tenant,
            replay_workflow_runs_request=replay_workflow_runs_request,
            _request_auth=_request_auth,
            _content_type=_content_type,
            _headers=_headers,
            _host_index=_host_index
        )

        _response_types_map: Dict[str, Optional[str]] = {
            '200': "ReplayWorkflowRunsResponse",
            '400': "APIErrors",
            '403': "APIErrors",
            '429': "APIErrors",
        }
        response_data = await self.api_client.call_api(
            *_param,
            _request_timeout=_request_timeout
        )
        await response_data.read()
        return self.api_client.response_deserialize(
            response_data=response_data,
            response_types_map=_response_types_map,
        ).data


    @validate_call
    async def workflow_run_update_replay_with_http_info(
        self,
        tenant: Annotated[str, Field(min_length=36, strict=True, max_length=36, description="The tenant id")],
        replay_workflow_runs_request: Annotated[ReplayWorkflowRunsRequest, Field(description="The workflow run ids to replay")],
        _request_timeout: Union[
            None,
            Annotated[StrictFloat, Field(gt=0)],
            Tuple[
                Annotated[StrictFloat, Field(gt=0)],
                Annotated[StrictFloat, Field(gt=0)]
            ]
        ] = None,
        _request_auth: Optional[Dict[StrictStr, Any]] = None,
        _content_type: Optional[StrictStr] = None,
        _headers: Optional[Dict[StrictStr, Any]] = None,
        _host_index: Annotated[StrictInt, Field(ge=0, le=0)] = 0,
    ) -> ApiResponse[ReplayWorkflowRunsResponse]:
        """Replay workflow runs

        Replays a list of workflow runs.

        :param tenant: The tenant id (required)
        :type tenant: str
        :param replay_workflow_runs_request: The workflow run ids to replay (required)
        :type replay_workflow_runs_request: ReplayWorkflowRunsRequest
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :type _request_timeout: int, tuple(int, int), optional
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the
                              authentication in the spec for a single request.
        :type _request_auth: dict, optional
        :param _content_type: force content-type for the request.
        :type _content_type: str, Optional
        :param _headers: set to override the headers for a single
                         request; this effectively ignores the headers
                         in the spec for a single request.
        :type _headers: dict, optional
        :param _host_index: set to override the host_index for a single
                            request; this effectively ignores the host_index
                            in the spec for a single request.
        :type _host_index: int, optional
        :return: Returns the result object.
        """ # noqa: E501

        _param = self._workflow_run_update_replay_serialize(
            tenant=tenant,
            replay_workflow_runs_request=replay_workflow_runs_request,
            _request_auth=_request_auth,
            _content_type=_content_type,
            _headers=_headers,
            _host_index=_host_index
        )

        _response_types_map: Dict[str, Optional[str]] = {
            '200': "ReplayWorkflowRunsResponse",
            '400': "APIErrors",
            '403': "APIErrors",
            '429': "APIErrors",
        }
        response_data = await self.api_client.call_api(
            *_param,
            _request_timeout=_request_timeout
        )
        await response_data.read()
        return self.api_client.response_deserialize(
            response_data=response_data,
            response_types_map=_response_types_map,
        )


    @validate_call
    async def workflow_run_update_replay_without_preload_content(
        self,
        tenant: Annotated[str, Field(min_length=36, strict=True, max_length=36, description="The tenant id")],
        replay_workflow_runs_request: Annotated[ReplayWorkflowRunsRequest, Field(description="The workflow run ids to replay")],
        _request_timeout: Union[
            None,
            Annotated[StrictFloat, Field(gt=0)],
            Tuple[
                Annotated[StrictFloat, Field(gt=0)],
                Annotated[StrictFloat, Field(gt=0)]
            ]
        ] = None,
        _request_auth: Optional[Dict[StrictStr, Any]] = None,
        _content_type: Optional[StrictStr] = None,
        _headers: Optional[Dict[StrictStr, Any]] = None,
        _host_index: Annotated[StrictInt, Field(ge=0, le=0)] = 0,
    ) -> RESTResponseType:
        """Replay workflow runs

        Replays a list of workflow runs.

        :param tenant: The tenant id (required)
        :type tenant: str
        :param replay_workflow_runs_request: The workflow run ids to replay (required)
        :type replay_workflow_runs_request: ReplayWorkflowRunsRequest
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :type _request_timeout: int, tuple(int, int), optional
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the
                              authentication in the spec for a single request.
        :type _request_auth: dict, optional
        :param _content_type: force content-type for the request.
        :type _content_type: str, Optional
        :param _headers: set to override the headers for a single
                         request; this effectively ignores the headers
                         in the spec for a single request.
        :type _headers: dict, optional
        :param _host_index: set to override the host_index for a single
                            request; this effectively ignores the host_index
                            in the spec for a single request.
        :type _host_index: int, optional
        :return: Returns the result object.
        """ # noqa: E501

        _param = self._workflow_run_update_replay_serialize(
            tenant=tenant,
            replay_workflow_runs_request=replay_workflow_runs_request,
            _request_auth=_request_auth,
            _content_type=_content_type,
            _headers=_headers,
            _host_index=_host_index
        )

        _response_types_map: Dict[str, Optional[str]] = {
            '200': "ReplayWorkflowRunsResponse",
            '400': "APIErrors",
            '403': "APIErrors",
            '429': "APIErrors",
        }
        response_data = await self.api_client.call_api(
            *_param,
            _request_timeout=_request_timeout
        )
        return response_data.response


    def _workflow_run_update_replay_serialize(
        self,
        tenant,
        replay_workflow_runs_request,
        _request_auth,
        _content_type,
        _headers,
        _host_index,
    ) -> RequestSerialized:

        _host = None

        _collection_formats: Dict[str, str] = {
        }

        _path_params: Dict[str, str] = {}
        _query_params: List[Tuple[str, str]] = []
        _header_params: Dict[str, Optional[str]] = _headers or {}
        _form_params: List[Tuple[str, str]] = []
        _files: Dict[
            str, Union[str, bytes, List[str], List[bytes], List[Tuple[str, bytes]]]
        ] = {}
        _body_params: Optional[bytes] = None

        # process the path parameters
        if tenant is not None:
            _path_params['tenant'] = tenant
        # process the query parameters
        # process the header parameters
        # process the form parameters
        # process the body parameter
        if replay_workflow_runs_request is not None:
            _body_params = replay_workflow_runs_request


        # set the HTTP header `Accept`
        if 'Accept' not in _header_params:
            _header_params['Accept'] = self.api_client.select_header_accept(
                [
                    'application/json'
                ]
            )

        # set the HTTP header `Content-Type`
        if _content_type:
            _header_params['Content-Type'] = _content_type
        else:
            _default_content_type = (
                self.api_client.select_header_content_type(
                    [
                        'application/json'
                    ]
                )
            )
            if _default_content_type is not None:
                _header_params['Content-Type'] = _default_content_type

        # authentication setting
        _auth_settings: List[str] = [
            'basicAuth', 
            'cookieAuth', 
            'bearerAuth'
        ]

        return self.api_client.param_serialize(
            method='POST',
            resource_path='/api/v1/tenants/{tenant}/workflow-runs/replay',
            path_params=_path_params,
            query_params=_query_params,
            header_params=_header_params,
            body=_body_params,
            post_params=_form_params,
            files=_files,
            auth_settings=_auth_settings,
            collection_formats=_collection_formats,
            _host=_host,
            _request_auth=_request_auth
        )





"""
Relative to this file I will have a prompt directory its located ../prompts
In this directory there will be a techniques directory and a directory for each model - gpt-3.5-turbo gpt-4, llama-2-70B, code-llama-7B etc

Each directory will have jinga2 templates for the prompts.
prompts in the model directories can use the techniques in the techniques directory.

Write the code I'd need to load and populate the templates.

I want the following functions:

class PromptEngine:

    def __init__(self, model):
        pass

    def load_prompt(model, prompt_name, prompt_ags) -> str:
        pass
"""

import glob
import os
from difflib import get_close_matches
from pathlib import Path
from typing import Any, List

import structlog
from jinja2 import Environment, FileSystemLoader

from mtmai.constants import SKYVERN_DIR

LOG = structlog.get_logger()


class PromptEngine:
    """
    Class to handle loading and populating Jinja2 templates for prompts.
    """

    def __init__(
        self, model: str, prompts_dir: Path = SKYVERN_DIR / "forge" / "prompts"
    ) -> None:
        """
        Initialize the PromptEngine with the specified model.

        Args:
            model (str): The model to use for loading prompts.
        """
        self.model = model

        try:
            # Get the list of all model directories
            models_dir = os.path.abspath(prompts_dir)
            model_names = [
                os.path.basename(os.path.normpath(d))
                for d in glob.glob(os.path.join(models_dir, "*/"))
                if os.path.isdir(d) and "techniques" not in d
            ]

            self.model = self.get_closest_match(self.model, model_names)

            self.env = Environment(loader=FileSystemLoader(models_dir))
        except Exception:
            LOG.error("Error initializing PromptEngine.", model=model, exc_info=True)
            raise

    @staticmethod
    def get_closest_match(target: str, model_dirs: List[str]) -> str:
        """
        Find the closest match to the target in the list of model directories.

        Args:
            target (str): The target model.
            model_dirs (list): The list of available model directories.

        Returns:
            str: The closest match to the target.
        """
        try:
            matches = get_close_matches(target, model_dirs, n=1, cutoff=0.1)
            return matches[0]
        except Exception:
            LOG.error(
                "Failed to get closest match.",
                target=target,
                model_dirs=model_dirs,
                exc_info=True,
            )
            raise

    def load_prompt(self, template: str, **kwargs: Any) -> str:
        """
        Load and populate the specified template.

        Args:
            template (str): The name of the template to load.
            **kwargs: The arguments to populate the template with.

        Returns:
            str: The populated template.
        """
        try:
            template = os.path.join(self.model, template)
            jinja_template = self.env.get_template(f"{template}.j2")
            return jinja_template.render(**kwargs)
        except Exception:
            LOG.error(
                "Failed to load prompt.",
                template=template,
                kwargs_keys=kwargs.keys(),
                exc_info=True,
            )
            raise

    def load_prompt_from_string(self, template: str, **kwargs: Any) -> str:
        """
        Load and populate the specified template from a string.

        Args:
            template (str): The template string to load.
            **kwargs: The arguments to populate the template with.

        Returns:
            str: The populated template.
        """
        try:
            jinja_template = self.env.from_string(template)
            return jinja_template.render(**kwargs)
        except Exception:
            LOG.error(
                "Failed to load prompt from string.",
                template=template,
                kwargs_keys=kwargs.keys(),
                exc_info=True,
            )
            raise



import uuid
from typing import Any

import structlog
from fastapi import HTTPException

# from mtmai.db.db import get_async_session
from sqlmodel import select
from sqlmodel.ext.asyncio.session import AsyncSession

from mtmai.core.security import get_password_hash, verify_password
from mtmai.models.models import User, UserCreate, UserUpdate

LOG = structlog.get_logger()


def _clean_result(obj):
    """Recursively change UUID -> str and serialize dictionaries"""
    if isinstance(obj, dict):
        return {k: _clean_result(v) for k, v in obj.items()}
    elif isinstance(obj, list):
        return [_clean_result(item) for item in obj]
    elif isinstance(obj, uuid.UUID):
        return str(obj)
    return obj


# async def execute_sql(
#     query: str, parameters: dict
# ) -> Union[List[Dict[str, Any]], int, None]:
#     parameterized_query = text(query)
#     async with get_async_session() as session:
#         try:
#             # await session.begin()
#             result = await session.exec(parameterized_query, parameters)
#             # await session.commit()
#             if result.returns_rows:
#                 json_result = [dict(row._mapping) for row in result.fetchall()]
#                 clean_json_result = _clean_result(json_result)
#                 return clean_json_result
#             else:
#                 return result.rowcount
#         except SQLAlchemyError as e:
#             await session.rollback()
#             LOG.warn(f"An error occurred: {e}")
#             return None
#         except Exception as e:
#             await session.rollback()
#             LOG.warn(f"An unexpected error occurred: {e}")
#             return None


# @cached(ttl=300)
# async def get_user_by_id2(user_id: str) -> User | None:
#     if not user_id:
#         return None
#     async with get_async_session() as session:
#         statement = select(User).where(User.id == user_id)
#         result = await session.exec(statement)
#         return result.first()


# async def get_organization_by_user_id(user_id: str | uuid.UUID):
#     from mtmai.forge.app import DATABASE

#     user = await get_user_by_id2(user_id)
#     organization_id = user.organization_id

#     organization = await DATABASE.get_organization(organization_id)
#     return organization


async def register_user(session: AsyncSession, user_in: UserCreate) -> User:
    from mtmai.forge.app import DATABASE

    new_organization = await DATABASE.create_organization(
        organization_name=user_in.username,
        webhook_callback_url="",
        max_steps_per_run="100",
        max_retries_per_step="3",
        domain="",
    )

    user = await get_user_by_email(session=session, email=user_in.email)
    if user:
        raise HTTPException(
            status_code=400,
            detail="The user with this email already exists in the system",
        )
    user_create = UserCreate.model_validate(user_in)
    user_create.organization_id = new_organization.organization_id
    user = await create_user(session=session, user_create=user_create)

    return user


async def create_user(*, session: AsyncSession, user_create: UserCreate) -> User:
    db_obj = User.model_validate(
        user_create, update={"hashed_password": get_password_hash(user_create.password)}
    )
    session.add(db_obj)
    await session.commit()
    await session.refresh(db_obj)
    return db_obj


async def update_user(
    *, session: AsyncSession, db_user: User, user_in: UserUpdate
) -> Any:
    user_data = user_in.model_dump(exclude_unset=True)
    extra_data = {}
    if "password" in user_data:
        password = user_data["password"]
        hashed_password = get_password_hash(password)
        extra_data["hashed_password"] = hashed_password
    db_user.sqlmodel_update(user_data, update=extra_data)
    session.add(db_user)
    await session.commit()
    await session.refresh(db_user)
    return db_user


async def get_user_by_email(*, session: AsyncSession, email: str) -> User | None:
    statement = select(User).where(User.email == email)
    result = await session.exec(statement)
    session_user = result.first()
    return session_user


async def get_user_by_username(*, session: AsyncSession, username: str) -> User | None:
    statement = select(User).where(User.username == username)
    result = await session.exec(statement)
    return result.first()


async def get_user_by_id(*, session: AsyncSession, id: str) -> User | None:
    if not id:
        return None
    statement = select(User).where(User.id == id)
    result = await session.exec(statement)
    return result.first()


async def authenticate(
    *, session: AsyncSession, email: str, password: str
) -> User | None:
    db_user = await get_user_by_email(session=session, email=email)
    if not db_user:
        return None
    if not verify_password(password, db_user.hashed_password):
        return None
    return db_user


# async def create_item(
#     *, session: AsyncSession, item_in: ItemCreate, owner_id: str
# ) -> Item:
#     db_item = Item.model_validate(item_in, update={"owner_id": owner_id})
#     session.add(db_item)
#     await session.commit()
#     await session.refresh(db_item)
#     return db_item



from langchain_core.prompts import ChatPromptTemplate

from mtmai.core.logging import get_logger
from mtmai.llm.llm import get_llm_long_context_default
from langchain_core.messages import AIMessage, HumanMessage, ToolMessage
from langchain_core.prompts import ChatPromptTemplate, MessagesPlaceholder
from langchain_core.runnables import RunnableConfig, RunnableLambda
from langchain_core.runnables import chain as as_runnable
from langchain_core.runnables import Runnable, RunnableConfig

from langgraph.graph import END, START, StateGraph
from pydantic import BaseModel, Field
from mtmai.agents.retrivers.web_search import search_engine
from mtmai.core.logging import get_logger

from mtmai.llm.llm import get_fast_llm
from mtmai.models.graph_config import Outline, ResearchState

logger = get_logger()

def format_conversation(interview_state):
    messages = interview_state["messages"]
    convo = "\n".join(f"{m.name}: {m.content}" for m in messages)
    return f'Conversation with {interview_state["editor"].name}\n\n' + convo


refine_outline_prompt = ChatPromptTemplate.from_messages(
    [
        (
            "system",
            """You are a Wikipedia writer. You have gathered information from experts and search engines. Now, you are refining the outline of the Wikipedia page. \
You need to make sure that the outline is comprehensive and specific. \
Topic you are writing about: {topic}

Old outline:

{old_outline}""",
        ),
        (
            "user",
            "Refine the outline based on your conversations with subject-matter experts:\n\nConversations:\n\n{conversations}\n\nWrite the refined Wikipedia outline:",
        ),
    ]
)



class RefineOutlineNode:
    def __init__(self, runnable: Runnable):
        self.runnable = runnable

    def node_name(self):
        return "refine_outline"

    async def __call__(self, state: ResearchState, config: RunnableConfig):
        """根据前面多个专家给出的观点，重新编辑大纲"""
        logger.info(f"进入 refine_outline node: {state}")
        convos = "\n\n".join(
            [
                format_conversation(interview_state)
                for interview_state in state["interview_results"]
            ]
        )

        # Using turbo preview since the context can get quite long
        refine_outline_chain = (
            refine_outline_prompt
            | get_llm_long_context_default().with_structured_output(Outline)
        )
        updated_outline = await refine_outline_chain.ainvoke(
            {
                "topic": state["topic"],
                "old_outline": state["outline"].as_str,
                "conversations": convos,
            }
        )
        return {**state, "outline": updated_outline}



from typing import Literal

from mtmai.core.config import settings


def get_langgraph_checkpointer(
    checkpointer_type: Literal["memory", "postgres"] = "memory",
):
    # if checkpointer_type == "postgres":
    #     return get_async_sqlite_checkpointer()
    from langgraph.checkpoint.memory import MemorySaver

    memory = MemorySaver()
    return memory


def get_async_postgres_checkpointer():
    from langgraph.checkpoint.postgres.aio import AsyncPostgresSaver
    from psycopg_pool import AsyncConnectionPool

    connection_kwargs = {
        "autocommit": True,
        "prepare_threshold": 0,
    }
    poll2 = AsyncConnectionPool(
        # Example configuration
        conninfo=settings.MTMAI_DATABASE_URL,
        max_size=20,
        kwargs=connection_kwargs,
    )
    # conn = poll2.connection()
    checkpointer = AsyncPostgresSaver(poll2)
    return checkpointer


def get_postgres_checkpointer():
    from langgraph.checkpoint.postgres import PostgresSaver
    from psycopg_pool import ConnectionPool

    connection_kwargs = {
        "autocommit": True,
        "prepare_threshold": 0,
    }
    poll2 = ConnectionPool(
        # Example configuration
        conninfo=settings.MTMAI_DATABASE_URL,
        max_size=20,
        kwargs=connection_kwargs,
    )
    # conn = poll2.connection()
    checkpointer = PostgresSaver(poll2)

    checkpointer.setup()

    return checkpointer



from phi.agent import Agent

from mtmai.agents.phiagents.config import model
from mtmai.agents.phiagents.finance_agent import finance_agent
from mtmai.agents.phiagents.web_agent import web_agent
from mtmai.core.logging import get_logger

logger = get_logger()
agent_team = Agent(
    team=[web_agent, finance_agent],
    show_tool_calls=True,
    markdown=True,
    model=model,
    monitoring=True,
    add_history_to_messages=True,
    read_chat_history=True,
    # debug_mode=True,
)



import json

from langchain_core.messages import AIMessage, HumanMessage, ToolMessage
from langchain_core.prompts import ChatPromptTemplate
from langchain_core.runnables import RunnableConfig
from mtmaisdk.clients.rest.models import AssisantState

from mtmai.agents.ctx import mtmai_context


class JokeWriterNode:
    def __init__(self):
        pass

    def node_name(self):
        return "joke_writer"

    async def __call__(self, state: AssisantState, config: RunnableConfig):
        # topic = state.topic
        self.state = state

        # if not topic:
        #     return {
        #         "error": "topic required",
        #     }

        write_joke_result = await self.write_joke_article("hello-topic")

        return {
            # **state,
            "messages": [
                AIMessage(
                    content=json.dumps(write_joke_result),
                )
            ],
        }

    async def write_joke_article(self, topic: str):
        """初始化大纲"""
        # ctx = get_mtmai_ctx()
        # parser = PydanticOutputParser(pydantic_object=Outline)
        direct_gen_outline_prompt = ChatPromptTemplate.from_messages(
            [
                (
                    "system",
                    "You are a witty and humorous joke generator. Please create a joke based on the given topic. The joke should be amusing and elicit laughter. "
                    "[ IMPORTANT ]:"
                    "\n- 必须使用简体中文"
                    "\n- Content length should be suitable for mobile device screens"
                    "\n- The joke should be witty and humorous, provoking laughter"
                    "\n- The content should be positive and uplifting, avoiding vulgarity"
                    "\n- The joke should be wholesome, free from violence, sexual content, or other inappropriate themes"
                    "\n- Ensure the joke is family-friendly and suitable for all audiences",
                ),
                ("user", "{topic}"),
            ]
        )
        ai_response = await mtmai_context.ainvoke_model(
            tpl=direct_gen_outline_prompt,
            inputs={"topic": topic},
        )
        return {
            "joke_content": "test",
        }



from typing import Callable

from langchain_core.messages import AIMessage, HumanMessage, ToolMessage
from langchain_core.runnables import Runnable, RunnableLambda
from langgraph.prebuilt import ToolNode
from pydantic import BaseModel

# from mtmai.models.graph_config import HomeChatState


def handle_tool_error(state) -> dict:
    error = state.get("error")
    tool_calls = state.messages[-1].tool_calls
    return {
        "messages": [
            ToolMessage(
                content=f"Error: {error!r}\n please fix your mistakes.",
                tool_call_id=tc["id"],
            )
            for tc in tool_calls
        ]
    }


def create_tool_node_with_fallback(tools: list):
    return ToolNode(tools).with_fallbacks(
        [RunnableLambda(handle_tool_error)],
        # exception_key="error"
    )


def _print_event(event: dict, _printed: set, max_length=1500):
    current_state = event.get("dialog_state")
    if current_state:
        print("Currently in: ", current_state[-1])
    message = event.get("messages")
    if message:
        if isinstance(message, list):
            message = message[-1]
        if message.id not in _printed:
            msg_repr = message.pretty_repr(html=True)
            if len(msg_repr) > max_length:
                msg_repr = msg_repr[:max_length] + " ... (truncated)"
            print(msg_repr)
            _printed.add(message.id)


def agent_node(state, agent, name):
    result = agent.invoke(state)
    return {
        "messages": [HumanMessage(content=result["messages"][-1].content, name=name)]
    }


# This node will be shared for exiting all specialized assistants
# def pop_dialog_state(state: HomeChatState) -> dict:
#     """Pop the dialog stack and return to the main assistant.

#     This lets the full graph explicitly track the dialog flow and delegate control
#     to specific sub-graphs.
#     """
#     messages = []
#     if state.messages[-1].tool_calls:
#         # Note: Doesn't currently handle the edge case where the llm performs parallel tool calls
#         messages.append(
#             ToolMessage(
#                 content="Resuming dialog with the host assistant. Please reflect on the past conversation and assist the user as needed.",
#                 tool_call_id=state.messages[-1].tool_calls[0]["id"],
#             )
#         )
#     return {
#         "dialog_state": "pop",
#         "messages": messages,
#     }


class CompleteOrEscalate(BaseModel):
    """A tool to mark the current task as completed and/or to escalate control of the dialog to the main assistant,
    who can re-route the dialog based on the user's needs."""

    cancel: bool = True
    reason: str

    class Config:
        json_schema_extra = {
            "example": {
                "cancel": True,
                "reason": "User changed their mind about the current task.",
            },
            "example 2": {
                "cancel": True,
                "reason": "I have fully completed the task.",
            },
            "example 3": {
                "cancel": False,
                "reason": "I need to search the user's emails or calendar for more information.",
            },
        }


async def ensure_valid_llm_response_v2(runnable: Runnable, messages: list):
    """
    确保 LLM 返回有效的回应
    原理：如果 LLM 返回了空内容，就主动模拟用户输入新的消息："Respond with a real output."
         催促 LLM 重新给出回应
    """

    messages_inner = list(messages)
    while True:
        ai_msg = await runnable.ainvoke(messages_inner)

        if isinstance(ai_msg, AIMessage):
            ai_msg_to_check = ai_msg.model_dump()
        else:
            ai_msg_to_check = ai_msg
        if (
            # ("parsed" not in ai_msg or not ai_msg["parsed"])
            ai_msg_to_check.get("parsed", None) is None
            and ai_msg_to_check.get("tool_calls", None) is not None
            and (
                ai_msg_to_check.get("content", None) is None
                or isinstance(ai_msg_to_check.get("content", None), list)
                and ai_msg_to_check.get("content", None)[0].get("text") is None
            )
        ):
            messages_inner = messages_inner + [("user", "Respond with a real output.")]
        else:
            break
    return ai_msg


# def create_entry_node(assistant_name: str, new_dialog_state: str) -> Callable:
#     """
#     Create a function to make an "entry" node for each workflow, stating "the current assistant ix assistant_name".
#     """

#     def entry_node(state: HomeChatState) -> dict:
#         tool_call_id = state.messages[-1].tool_calls[0]["id"]
#         return {
#             "messages": [
#                 ToolMessage(
#                     content=f"The assistant is now the {assistant_name}. Reflect on the above conversation between the host assistant and the user."
#                     f" The user's intent is unsatisfied. Use the provided tools to assist the user. Remember, you are {assistant_name},"
#                     " and the booking, update, other other action is not complete until after you have successfully invoked the appropriate tool."
#                     " If the user changes their mind or needs help for other tasks, call the CompleteOrEscalate function to let the primary host assistant take control."
#                     " Do not mention who you are - just act as the proxy for the assistant.",
#                     tool_call_id=tool_call_id,
#                 )
#             ],
#             "dialog_state": new_dialog_state,
#         }

#     return entry_node


internal_node_types = set(
    [
        "RunnableSequence",
        "RunnableLambda",
        "RunnableParallel<raw>",
        "RunnableWithFallbacks",
        "_write",
        "start",
        "end",
    ]
)
skip_kinds = set(
    [
        "on_tool_start",
        "on_tool_end",
        "on_chat_model_stream",
    ]
)


def is_internal_node(node_name: str):
    """
    判断是否是内部节点
    """
    return node_name in internal_node_types


def is_skip_kind(kind: str):
    """
    判断是否是跳过类型
    """
    return kind in skip_kinds



import re





def sanitize_filename(filename: str) -> str:
    """
    Sanitize a given filename by replacing characters that are invalid
    in Windows file paths with an underscore ('_').

    This function ensures that the filename is compatible with all
    operating systems by removing or replacing characters that are
    not allowed in Windows file paths. Specifically, it replaces
    the following characters: < > : " / \ | ? *

    Parameters:
    filename (str): The original filename to be sanitized.

    Returns:
    str: The sanitized filename with invalid characters replaced by an underscore.

    Examples:
    >>> sanitize_filename('invalid:file/name*example?.txt')
    'invalid_file_name_example_.txt'

    >>> sanitize_filename('valid_filename.txt')
    'valid_filename.txt'
    """
    return re.sub(r'[<>:"/\\|?*]', '_', filename)



from langchain_core.messages import AIMessage
from langchain_core.runnables import Runnable, RunnableConfig
from langgraph.graph import END, StateGraph

from mtmai.agents.nodes import gen_answer_node
from mtmai.agents.nodes.generate_question import GenQuestionNode
# from mtmai.agents.states.research_state import InterviewState, ResearchState
from mtmai.core.logging import get_logger
from mtmai.models.graph_config import InterviewState, ResearchState

logger = get_logger()

InterviewState
def route_messages(state: InterviewState, name: str = "Subject_Matter_Expert"):
    max_num_turns = 3
    messages = state["messages"]
    num_responses = len(
        [m for m in messages if isinstance(m, AIMessage) and m.name == name]
    )
    if num_responses >= max_num_turns:
        return END
    last_question = messages[-2]
    if last_question.content.endswith("Thank you so much for your help!"):
        return END
    return "continue"


# 旧版 (基于 langgraph)
class ConductInterviewNode:
    def __init__(
        self,
        runnable: Runnable,
    ):
        self.runnable = runnable

    def node_name(self):
        return "conduct_interview_node"

    async def __call__(self, state: ResearchState, config: RunnableConfig):
        topic = state["topic"]

        # 前面步骤确定了 问题解答专家的角色(editor), 现在动态构建自动，并发对多个专家进行采访，看他们观点
        initial_states = [
            {
                "editor": editor,
                "messages": [
                    AIMessage(
                        content=f"So you said you were writing an article on {topic}?",
                        name="Subject_Matter_Expert",
                    )
                ],
            }
            for editor in state["editors"]
        ]
        # We call in to the sub-graph here to parallelize the interviews

        builder = StateGraph(InterviewState)

        builder.add_node("ask_question", GenQuestionNode(self.runnable))
        builder.add_node(
            "answer_question",
            gen_answer_node.GenAnswerNode(
                self.runnable, name="Subject_Matter_Expert", max_str_len=15000
            ),
        )
        builder.add_conditional_edges(
            "answer_question",
            route_messages,
            {
                "continue": "ask_question",
                "error": END,
                END: END,
            },
        )
        builder.add_edge("ask_question", "answer_question")

        builder.set_entry_point("ask_question")
        interview_graph = builder.compile().with_config(run_name="Conduct Interviews")

        logger.info(f"开始进行 采访, 数量: {len(initial_states)}")

        enabled_batch = False
        if enabled_batch:
            interview_results = await interview_graph.abatch(initial_states, config)
        else:
            interview_results = []
            first_state = initial_states[0]
            # for a in initial_states:
            #     interview_results.append(await interview_graph.ainvoke(a))
            # 暂时仅处理第一个
            interview_results.append(await interview_graph.ainvoke(first_state))
        # logger.info(f"采访结束, 数量: {len(interview_results)}")
        return {
            **state,
            "interview_results": interview_results,
        }


class ConductInterviewNodeV2:
    def __init__(
        self,
        runnable: Runnable,
    ):
        self.runnable = runnable

    def node_name(self):
        return "conduct_interview_node"

    async def __call__(self, state: ResearchState):
        topic = state["topic"]

        # 前面步骤确定了 问题解答专家的角色(editor), 现在动态构建自动，并发对多个专家进行采访，看他们观点
        initial_states = [
            {
                "editor": editor,
                "messages": [
                    AIMessage(
                        content=f"So you said you were writing an article on {topic}?",
                        name="Subject_Matter_Expert",
                    )
                ],
            }
            for editor in state["editors"]
        ]
        # We call in to the sub-graph here to parallelize the interviews

        builder = StateGraph(InterviewState)

        builder.add_node("ask_question", GenQuestionNode(self.runnable))
        builder.add_node(
            "answer_question",
            gen_answer_node.GenAnswerNode(
                self.runnable, name="Subject_Matter_Expert", max_str_len=15000
            ),
        )
        builder.add_conditional_edges(
            "answer_question",
            route_messages,
            {
                "continue": "ask_question",
                "error": END,
                END: END,
            },
        )
        builder.add_edge("ask_question", "answer_question")

        builder.set_entry_point("ask_question")
        interview_graph = builder.compile().with_config(run_name="Conduct Interviews")

        logger.info(f"开始进行 采访, 数量: {len(initial_states)}")

        enabled_batch = False
        if enabled_batch:
            interview_results = await interview_graph.abatch(initial_states, config)
        else:
            interview_results = []
            first_state = initial_states[0]
            # for a in initial_states:
            #     interview_results.append(await interview_graph.ainvoke(a))
            # 暂时仅处理第一个
            interview_results.append(await interview_graph.ainvoke(first_state))
        # logger.info(f"采访结束, 数量: {len(interview_results)}")
        return {
            **state,
            "interview_results": interview_results,
        }



import structlog
from crewai.tools.base_tool import Tool
from mtmai.agents.tools.web_search import search_engine
from pydantic import BaseModel, Field

LOG = structlog.get_logger()


class ToFlightBookingAssistant(BaseModel):
    """Transfers work to a specialized assistant to handle flight updates and cancellations."""

    request: str = Field(
        description="Any necessary followup questions the update flight assistant should clarify before proceeding."
    )


class ToBookCarRental(BaseModel):
    """Transfers work to a specialized assistant to handle car rental bookings."""

    location: str = Field(
        description="The location where the user wants to rent a car."
    )
    start_date: str = Field(description="The start date of the car rental.")
    end_date: str = Field(description="The end date of the car rental.")
    request: str = Field(
        description="Any additional information or requests from the user regarding the car rental."
    )

    class Config:
        json_schema_extra = {
            "example": {
                "location": "Basel",
                "start_date": "2023-07-01",
                "end_date": "2023-07-05",
                "request": "I need a compact car with automatic transmission.",
            }
        }


def get_tools(tools_name: str | list[str]):
    """
    根据工具名称获取工具函数
    """
    tools = []
    if isinstance(tools_name, str):
        tools_name = [tools_name]

    for tool_name in tools_name:
        if tool_name == "search_engine":
            # tools.append(Tool.from_langchain(search_engine_tool))
            tools.append(Tool.from_langchain(search_engine))
    return tools



import logging
from typing import Annotated, AsyncGenerator, Optional

from fastapi import Depends, HTTPException, Request, status
from fastapi.security import OAuth2PasswordBearer
from langgraph.checkpoint.postgres.aio import AsyncPostgresSaver
from sqlmodel import Session
from sqlmodel.ext.asyncio.session import AsyncSession

from mtmai.core.config import HEADER_SITE_HOST, settings
from mtmai.crud.curd_site import get_site_domain
from mtmai.db.db import get_async_engine, get_checkpointer
from mtmai.models.models import User
from mtmai.models.site import Site

from .database import DatabaseManager
from .teammanager import TeamManager
from .ws_connection import WebSocketManager

logger = logging.getLogger(__name__)


# Global manager instances
_db_manager: Optional[DatabaseManager] = None
_websocket_manager: Optional[WebSocketManager] = None
_team_manager: Optional[TeamManager] = None


reusable_oauth2 = OAuth2PasswordBearer(
    tokenUrl=f"{settings.API_V1_STR}/login/access-token",
    auto_error=False,  # 没有 token header 时不触发异常
)


# def get_db() -> Generator[Session, None, None]:
#     with Session(engine) as session:
#         yield session


async def get_db() -> DatabaseManager:
    """Dependency provider for database manager"""
    if not _db_manager:
        raise HTTPException(
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
            detail="Database manager not initialized",
        )
    return _db_manager


async def get_websocket_manager() -> WebSocketManager:
    """Dependency provider for connection manager"""
    if not _websocket_manager:
        raise HTTPException(
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
            detail="Connection manager not initialized",
        )
    return _websocket_manager


async def get_team_manager() -> TeamManager:
    """Dependency provider for team manager"""
    if not _team_manager:
        raise HTTPException(
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
            detail="Team manager not initialized",
        )
    return _team_manager


# Authentication dependency


async def get_current_user(
    # Add your authentication logic here
    # For example: token: str = Depends(oauth2_scheme)
) -> str:
    """
    Dependency for getting the current authenticated user.
    Replace with your actual authentication logic.
    """
    # Implement your user authentication here
    return "user_id"  # Replace with actual user identification


# Manager initialization and cleanup


async def init_managers(database_uri: str, config_dir: str, app_root: str) -> None:
    """Initialize all manager instances"""
    global _db_manager, _websocket_manager, _team_manager

    logger.info("Initializing managers...")

    try:
        # Initialize database manager
        _db_manager = DatabaseManager(engine_uri=database_uri, base_dir=app_root)
        _db_manager.initialize_database(auto_upgrade=settings.UPGRADE_DATABASE)

        # init default team config
        await _db_manager.import_teams_from_directory(
            config_dir, settings.DEFAULT_USER_ID, check_exists=True
        )

        # Initialize connection manager
        _websocket_manager = WebSocketManager(db_manager=_db_manager)
        logger.info("Connection manager initialized")

        # Initialize team manager
        _team_manager = TeamManager()
        logger.info("Team manager initialized")

    except Exception as e:
        logger.error(f"Failed to initialize managers: {str(e)}")
        await cleanup_managers()  # Cleanup any partially initialized managers
        raise


async def cleanup_managers() -> None:
    """Cleanup and shutdown all manager instances"""
    global _db_manager, _websocket_manager, _team_manager

    logger.info("Cleaning up managers...")

    # Cleanup connection manager first to ensure all active connections are closed
    if _websocket_manager:
        try:
            await _websocket_manager.cleanup()
        except Exception as e:
            logger.error(f"Error cleaning up connection manager: {str(e)}")
        finally:
            _websocket_manager = None

    # TeamManager doesn't need explicit cleanup since WebSocketManager handles it
    _team_manager = None

    # Cleanup database manager last
    if _db_manager:
        try:
            await _db_manager.close()
        except Exception as e:
            logger.error(f"Error cleaning up database manager: {str(e)}")
        finally:
            _db_manager = None

    logger.info("All managers cleaned up")


# Utility functions for dependency management


def get_manager_status() -> dict:
    """Get the initialization status of all managers"""
    return {
        "database_manager": _db_manager is not None,
        "websocket_manager": _websocket_manager is not None,
        "team_manager": _team_manager is not None,
    }


# Combined dependencies


async def get_managers():
    """Get all managers in one dependency"""
    return {
        "db": await get_db(),
        "connection": await get_websocket_manager(),
        "team": await get_team_manager(),
    }


# Error handling for manager operations


class ManagerOperationError(Exception):
    """Custom exception for manager operation errors"""

    def __init__(self, manager_name: str, operation: str, detail: str):
        self.manager_name = manager_name
        self.operation = operation
        self.detail = detail
        super().__init__(f"{manager_name} failed during {operation}: {detail}")


# Dependency for requiring specific managers


def require_managers(*manager_names: str):
    """Decorator to require specific managers for a route"""

    async def dependency():
        status = get_manager_status()
        missing = [name for name in manager_names if not status.get(f"{name}_manager")]
        if missing:
            raise HTTPException(
                status_code=status.HTTP_503_SERVICE_UNAVAILABLE,
                detail=f"Required managers not available: {', '.join(missing)}",
            )
        return True

    return Depends(dependency)


async def get_asession() -> AsyncGenerator[AsyncSession, None]:
    async with AsyncSession(get_async_engine()) as session:
        yield session


SessionDep = Annotated[Session, Depends(get_db)]

AsyncSessionDep = Annotated[AsyncSession, Depends(get_asession)]


TokenDep = Annotated[str, Depends(reusable_oauth2)]


def get_host_from_request(request: Request):
    host = request.headers.get("Host")
    return host


HostDep = Annotated[str, Depends(get_host_from_request)]


# def get_current_user(session: SessionDep, token: TokenDep, request: Request) -> User:
#     token = token or request.cookies.get(settings.COOKIE_ACCESS_TOKEN)
#     try:
#         payload = jwt.decode(
#             token, settings.SECRET_KEY, algorithms=[security.ALGORITHM]
#         )
#         token_data = TokenPayload(**payload)
#     except (InvalidTokenError, ValidationError):
#         raise HTTPException(
#             status_code=status.HTTP_403_FORBIDDEN,
#             detail="Could not validate credentials",
#         )
#     user = session.get(User, token_data.sub)
#     if not user:
#         raise HTTPException(status_code=404, detail="User not found")
#     if not user.is_active:
#         raise HTTPException(status_code=400, detail="Inactive user")
#     return user


CurrentUser = Annotated[User, Depends(get_current_user)]


def get_current_active_superuser(current_user: CurrentUser) -> User:
    if not current_user.is_superuser:
        raise HTTPException(
            status_code=403, detail="The user doesn't have enough privileges"
        )
    return current_user


def get_optional_current_user(
    session: SessionDep, token: TokenDep, request: Request
) -> User | None:
    token = token or request.cookies.get(settings.COOKIE_ACCESS_TOKEN)
    if not token:
        return None
    try:
        return get_current_user(session, token, request)
    except HTTPException:
        return None


OptionalUserDep = Annotated[User | None, Depends(get_optional_current_user)]


CheckPointerDep = Annotated[AsyncPostgresSaver, Depends(get_checkpointer)]


async def get_site(session: AsyncSessionDep, request: Request) -> Site:
    """
    根据入站域名获取对应是site 对象
    """
    income_domain = request.headers.get(HEADER_SITE_HOST)
    if not income_domain:
        # 尝试从多个来源获取前端域名
        # 1. 检查反向代理头
        if "X-Forwarded-Host" in request.headers:
            income_domain = request.headers["X-Forwarded-Host"]
        # 2. 检查 Referer 头
        elif "Referer" in request.headers:
            from urllib.parse import urlparse

            referer = request.headers["Referer"]
            income_domain = urlparse(referer).netloc
        # 3. 检查 Origin 头
        elif "Origin" in request.headers:
            from urllib.parse import urlparse

            origin = request.headers["Origin"]
            income_domain = urlparse(origin).netloc
        # 4. 如果以上都失败，使用 Host 头作为后备
        else:
            income_domain = request.headers.get("Host")

    if income_domain:
        site = await get_site_domain(session, income_domain)
        return site
    else:
        raise HTTPException(status_code=400, detail="Unable to determine site domain")


SiteDep = Annotated[Site, Depends(get_site)]
SiteDep = Annotated[Site, Depends(get_site)]



from contextlib import contextmanager

from sqlmodel import Session


@contextmanager
def transaction_scope(db: Session):
    """Context manager for handling database transactions."""
    try:
        yield
        db.commit()
    except Exception as e:
        db.rollback()
        raise e



import logging
import operator
from collections.abc import Sequence
from functools import lru_cache
from typing import Annotated

from langgraph.constants import Send
from langgraph.graph import END, START, StateGraph
from langgraph.graph.state import CompiledStateGraph
from pydantic import BaseModel

from mtmai.llm.llm import get_llm_chatbot_default

logger = logging.getLogger(__name__)


class BlogWriterAgentState(BaseModel):
    subjects: list[str] | None = None
    jokes: Annotated[Sequence[str], operator.add] | None = (
        None  # = Field(sa_column=Column(JSON))
    )
    best_selected_joke: str | None = None
    # messages: Annotated[list, add_messages] = Field(sa_column=Column(JSON))
    messages: list[dict] | None = None  # = Field(sa_column=Column(JSON))
    ask_human: bool = False


subjects_prompt = """Generate a comma separated list of between 2 and 5 examples related to: {topic}."""
joke_prompt = """Generate a joke about {subject}"""
best_joke_prompt = """Below are a bunch of jokes about {topic}. Select the best one! Return the ID of the best one.

{jokes}"""


class Subjects(BaseModel):
    subjects: list[str]


class Joke(BaseModel):
    joke: str


class BestJoke(BaseModel):
    id: int


class JokeState(BaseModel):
    subject: str


# This is the function we will use to generate the subjects of the jokes
def generate_topics(state: BlogWriterAgentState):
    llm = get_llm_chatbot_default()
    latest_message = state.messages[-1]
    prompt = subjects_prompt.format(topic=latest_message["content"])
    response = llm.with_structured_output(Subjects).invoke(prompt)
    state.subjects = response.subjects
    # return {"subjects": response.subjects}
    return state


# Here we generate a joke, given a subject
def generate_joke(state: JokeState):
    llm = get_llm_chatbot_default()
    prompt = joke_prompt.format(subject=state.subject)
    response = llm.with_structured_output(Joke).invoke(prompt)
    # state.jokes = [response.joke]
    return {"jokes": [response.joke]}
    # return state


# Here we define the logic to map out over the generated subjects
# We will use this an edge in the graph
def continue_to_jokes(state: BlogWriterAgentState):
    # We will return a list of `Send` objects
    # Each `Send` object consists of the name of a node in the graph
    # as well as the state to send to that node
    send_list = []
    for s in state.subjects:
        joke_state = JokeState(
            subject=s,
        )
        send_list.append(Send("generate_joke", joke_state))
    # return [Send("generate_joke", ) for JokeAgentState(subjects=s) in state.subjects]
    return send_list


# Here we will judge the best joke
def best_joke(state: BlogWriterAgentState):
    llm = get_llm_chatbot_default()

    jokes = "\n\n".join(state.jokes)
    latest_message = state.messages[-1]

    prompt = best_joke_prompt.format(topic=latest_message["content"], jokes=jokes)
    response = llm.with_structured_output(BestJoke).invoke(prompt)

    idx = response.id

    # Ensure idx is within the bounds of the jokes list
    if idx >= len(state.jokes):
        idx = len(state.jokes) - 1

    return {"best_selected_joke": state.jokes[idx]}


@lru_cache
def get_workflow() -> CompiledStateGraph:
    graph = StateGraph(BlogWriterAgentState)
    graph.add_node("generate_topics", generate_topics)
    graph.add_node("generate_joke", generate_joke)
    graph.add_node("best_joke", best_joke)
    graph.add_edge(START, "generate_topics")
    graph.add_conditional_edges("generate_topics", continue_to_jokes, ["generate_joke"])
    graph.add_edge("generate_joke", "best_joke")
    graph.add_edge("best_joke", END)
    app = graph.compile()
    return app



import asyncio
import json

from sqlalchemy.ext.asyncio import AsyncEngine

from mtmai.core.logging import get_logger
from mtmai.mtlibs.mq.pq_queue import AsyncPGMQueue
from mtmai.mtlibs.mtmcrawler.mtmcrawler import MTCrawlerPageParams, _fetch_page_html
from mtmlib.queue.queue import Message

logger = get_logger()

mtm_crawl_queue = "mtm-crawl-queue"


class CrawlWorker:
    def __init__(self, mq: AsyncPGMQueue, engine: AsyncEngine):
        self.mq = mq
        self.db = engine
        self.is_running = False

    async def start(self):
        # logger.info("🕷️ 🟢 Start MTM crawler worker ")
        self.is_running = True

        asyncio.create_task(self._pull_messages())

    async def stop(self):
        logger.info("🕷️ 🛑 Stop MTM crawler worker ")
        self.is_running = False

    async def _pull_messages(self):
        logger.info(f"🕷️ 🟢 pull_messages from {mtm_crawl_queue}")
        await self.mq.create_queue(queue=mtm_crawl_queue)
        while self.is_running:
            try:
                msg = await self.mq.read(queue=mtm_crawl_queue)
                if msg:
                    await self._handle_message(msg)
                else:
                    await asyncio.sleep(1)
            except Exception as e:
                logger.error(f"🕷️ 🔴 pull_messages error: {e}")
                await asyncio.sleep(1)

    async def _handle_message(self, msg: Message):
        # logger.info(f"读取到消息队列的消息 : {msg.msg_id}")
        params = MTCrawlerPageParams(**json.loads(msg.message))
        # logger.info(f"处理消息队列的消息 : {params.url}")
        await self.crawl_page(params)
        await self.mq.ack(mtm_crawl_queue, msg.msg_id)
        logger.debug(f"处理消息队完成{msg.msg_id}")

    async def enqueue_url(self, site_id: str, url: str):
        params = MTCrawlerPageParams(site_id=site_id, url=url, current_depth=0)
        message = json.dumps(params.model_dump())
        await self.mq.send(mtm_crawl_queue, message)
        logger.info(f"发送消息队列的消息 : {params.url} 完成")

    async def crawl_page(self, params: MTCrawlerPageParams):
        """
        爬取一个页面并建立页面索引
        """
        html = await _fetch_page_html(params.url)
        if html is None:
            return None
        logger.info(f"crawl_page: {len(html)}, {params.url}")

        # page_item = MTCrawlPage(
        #     site_id=params.site_id,
        #     url=params.url,
        #     depth=params.current_depth,
        #     title="",
        #     description="",
        #     keywords="",
        #     author="fake_author",
        #     copyright="fake_copyright",
        # )
        # async with AsyncSession(self.db) as session:
        #     await session.add(page_item)
        #     await session.commit()
        # return update_item
        return None



import logging

import numpy as np
import pandas as pd
import transformers
from sklearn.model_selection import train_test_split

from mtmtrain import dataset
from mtmtrain.ipy import ipyutils

logger = logging.getLogger()


def train():
    all_datasets = ["text_classify/bbc-text.csv"]

    logger.info("加载数据集 %s", all_datasets)
    for ds in all_datasets:
        dataset.load_dataset(ds)
    # Download the dataset and put it in subfolder called data
    datapath = dataset.get_dataset_path("text_classify/bbc-text.csv")
    df = pd.read_csv(datapath)
    df = df[["category", "text"]]

    logging.info("简单信息浏览")
    ipyutils.display(df.head())

    # Renaming, Input -> X, Output -> y
    X = df["text"]
    y = np.unique(df["category"], return_inverse=True)[1]
    logger.info("category 列 %s", y)
    print(y)

    logging.info("开始用 TensorFlow  Building a Text Classification")

    tokenizer = transformers.DistilBertTokenizer.from_pretrained(
        "distilbert-base-uncased"
    )
    X_tf = [
        tokenizer(text, padding="max_length", max_length=512, truncation=True)[
            "input_ids"
        ]
        for text in X
    ]
    X_tf = np.array(X_tf, dtype="int32")

    # Train/test split

    X_tf_train, X_tf_test, y_tf_train, y_tf_test = train_test_split(
        X_tf, y, test_size=0.3, random_state=42, stratify=y
    )
    print("Shape of training data: ", X_tf_train.shape)
    print("Shape of test data: ", X_tf_test.shape)

    logging.info("Build the Model")
    # Get BERT layer
    config = transformers.DistilBertConfig(dropout=0.2, attention_dropout=0.2)
    dbert_tf = transformers.TFDistilBertModel.from_pretrained(
        "distilbert-base-uncased", config=config, trainable=False
    )
    logger.info(dbert_tf)

    # Let's create a sample of size 5 from the training data
    sample = X_tf_train[0:5]
    print("Object type: ", type(dbert_tf(sample)))
    print("Output format (shape): ", dbert_tf(sample)[0].shape)
    print(
        "Output used as input for the classifier (shape): ",
        dbert_tf(sample)[0][:, 0, :].shape,
    )



from contextlib import asynccontextmanager

from prefect import get_run_logger

from mtmai.crud.curd_logs import LogItemCreateReq, create_log_item

log_type = "mtflow"


def get_flow_util():
    return FlowsUtil()


class FlowsUtil:
    def __init__(self):
        self.logger = get_run_logger()

    async def info(self, text: str):
        self.logger.info(text)
        await create_log_item(
            LogItemCreateReq(app=log_type, text=text, type=text, level=3)
        )

    async def error(self, text: str):
        self.logger.error(text)
        await create_log_item(
            LogItemCreateReq(app=log_type, text=text, type=text, level=1)
        )

    async def warn(self, text: str):
        self.logger.warning(text)
        await create_log_item(
            LogItemCreateReq(app=log_type, text=text, type=text, level=2)
        )


@asynccontextmanager
async def mtflow_util():
    # task_id = "demo_task_id123"
    flow_util = FlowsUtil()

    # mq = MtTaskMQ(task_id)
    # await mq.init()
    await flow_util.info("任务开始-----------------------------")
    try:
        yield flow_util
    finally:
        await flow_util.info("任务完成-----------------------------")



import asyncio

import litellm
from crewai import Agent, Crew, Process, Task
from fastapi.encoders import jsonable_encoder
# from mtmai.workflows.step_base import get_wf_log_callbacks
from mtmaisdk.clients.rest.models import CrewAIParams
from mtmaisdk.clients.rest.models.call_agent import CallAgent
from mtmaisdk.context.context import Context

from mtmai.worker import wfapp
from mtmai.workflows.crews import call_crew


@wfapp.workflow(
    on_events=["crewai:call"],
    input_validator=CrewAIParams,
)
class FlowCrewAIAgent:
    @wfapp.step(timeout="10m", retries=1)
    async def run(self, hatctx: Context):
        input = CallAgent.model_validate(hatctx.workflow_input())
        # callback = get_wf_log_callbacks(hatctx)
        # llm = get_llm(input.llm, callback)
        agents = []
        agent_dict = {}
        for agent in input.agents:
            new_agent = Agent(
                role=agent.role,
                backstory=agent.backstory,
                goal=agent.goal,
                llm=llm,
                verbose=True,
                max_retry_limit=100,
                max_rpm=60,
                step_callback=callback,
                task_callback=callback,
                tools=[
                    self.get_tool(self.ctx, "httpfetch"),
                    # self.get_tool(self.ctx, "runnewtask"),
                ],
                memory=True,
            )
            agents.append(new_agent)
            agent_dict[agent.name] = new_agent

        tasks = []

        for task in input.tasks:
            target_agent = agent_dict[task.agent]
            if not target_agent:
                raise ValueError(f"缺少名称为 {task.agent} 的 agent")
            new_task = Task(
                description=task.description,
                expected_output=task.expected_output,
                agent=target_agent,
                callback=callback,
                # output_pydantic=getOuputSchemaByName(task.output_json_schema_name)
                # if task.output_json_schema_name is not None
                # else None,
            )
            tasks.append(new_task)

        crew = Crew(
            agents=agents,
            tasks=tasks,
            process=Process.sequential,
            verbose=True,
            step_callback=callback,
            task_callback=callback,
            memory=True,
            knowledge_sources=[self.get_crew_knownledge()],
            embedder_config={
                "provider": "ollama",
                "config": {"model": "nomic-embed-text:latest"},
            },
        )
        crew.short_term_memory = self.get_crew_short_term_mem(crew)
        crew.long_term_memory = self.get_crew_long_term_mem(crew)
        crew.entity_memory = self.get_crew_entiry_mem(crew)

        try:
            result = jsonable_encoder(await call_crew(crew, input.input))
            if isinstance(result, str):
                # self.ctx.put_stream('0:"hello from crew result"\n')
                self.emit(result)
                return {"raw": result}
            return result
        except litellm.RateLimitError as e:
            await asyncio.sleep(20)
            self.ctx.log("速率限制，休眠")
            raise e
            self.ctx.log("速率限制，休眠")
            raise e



from typing import List

from selenium.webdriver.common.by import By
import selenium.webdriver.remote.webelement


class WebElement(selenium.webdriver.remote.webelement.WebElement):
    def click_safe(self):
        super().click()
        self._parent.reconnect(0.1)

    def children(
        self, tag=None, recursive=False
    ) -> List[selenium.webdriver.remote.webelement.WebElement]:
        """
        returns direct child elements of current element
        :param tag: str,  if supplied, returns <tag> nodes only
        """
        script = "return [... arguments[0].children]"
        if tag:
            script += ".filter( node => node.tagName === '%s')" % tag.upper()
        if recursive:
            return list(_recursive_children(self, tag))
        return list(self._parent.execute_script(script, self))


class UCWebElement(WebElement):
    """
    Custom WebElement class which makes it easier to view elements when
    working in an interactive environment.

    standard webelement repr:
    <selenium.webdriver.remote.webelement.WebElement (session="85ff0f671512fa535630e71ee951b1f2", element="6357cb55-92c3-4c0f-9416-b174f9c1b8c4")>

    using this WebElement class:
    <WebElement(<a class="mobile-show-inline-block mc-update-infos init-ok" href="#" id="main-cat-switcher-mobile">)>

    """

    def __init__(self, parent, id_):
        super().__init__(parent, id_)
        self._attrs = None

    @property
    def attrs(self):
        if not self._attrs:
            self._attrs = self._parent.execute_script(
                """
                var items = {};
                for (index = 0; index < arguments[0].attributes.length; ++index)
                {
                 items[arguments[0].attributes[index].name] = arguments[0].attributes[index].value
                };
                return items;
                """,
                self,
            )
        return self._attrs

    def __repr__(self):
        strattrs = " ".join([f'{k}="{v}"' for k, v in self.attrs.items()])
        if strattrs:
            strattrs = " " + strattrs
        return f"{self.__class__.__name__} <{self.tag_name}{strattrs}>"


def _recursive_children(element, tag: str = None, _results=None):
    """
    returns all children of <element> recursively

    :param element: `WebElement` object.
            find children below this <element>

    :param tag: str = None.
            if provided, return only <tag> elements. example: 'a', or 'img'
    :param _results: do not use!
    """
    results = _results or set()
    for element in element.children():
        if tag:
            if element.tag_name == tag:
                results.add(element)
        else:
            results.add(element)
        results |= _recursive_children(element, tag, results)
    return results



# api/ws.py
import asyncio
import json
from datetime import datetime
from uuid import UUID

from fastapi import APIRouter, Depends, WebSocket, WebSocketDisconnect
from loguru import logger

from ..deps import get_db, get_websocket_manager
from ..models.ag_db import Run, RunStatus
from ..ws_connection import WebSocketManager

router = APIRouter()


@router.websocket("/runs/{run_id}")
async def run_websocket(
    websocket: WebSocket,
    run_id: UUID,
    ws_manager: WebSocketManager = Depends(get_websocket_manager),
    db=Depends(get_db),
):
    """WebSocket endpoint for run communication"""
    # Verify run exists and is in valid state
    run_response = db.get(Run, filters={"id": run_id}, return_json=False)
    if not run_response.status or not run_response.data:
        logger.warning(f"Run not found: {run_id}")
        await websocket.close(code=4004, reason="Run not found")
        return

    run = run_response.data[0]
    if run.status not in [RunStatus.CREATED, RunStatus.ACTIVE]:
        await websocket.close(code=4003, reason="Run not in valid state")
        return

    # Connect websocket
    connected = await ws_manager.connect(websocket, run_id)
    if not connected:
        await websocket.close(code=4002, reason="Failed to establish connection")
        return

    try:
        logger.info(f"WebSocket connection established for run {run_id}")

        while True:
            try:
                raw_message = await websocket.receive_text()
                message = json.loads(raw_message)

                if message.get("type") == "start":
                    # Handle start message
                    logger.info(f"Received start request for run {run_id}")
                    task = message.get("task")
                    team_config = message.get("team_config")
                    if task and team_config:
                        # await ws_manager.start_stream(run_id, task, team_config)
                        asyncio.create_task(
                            ws_manager.start_stream(run_id, task, team_config)
                        )
                    else:
                        logger.warning(f"Invalid start message format for run {run_id}")
                        await websocket.send_json(
                            {
                                "type": "error",
                                "error": "Invalid start message format",
                                "timestamp": datetime.utcnow().isoformat(),
                            }
                        )

                elif message.get("type") == "stop":
                    logger.info(f"Received stop request for run {run_id}")
                    reason = message.get("reason") or "User requested stop/cancellation"
                    await ws_manager.stop_run(run_id, reason=reason)
                    break

                elif message.get("type") == "ping":
                    await websocket.send_json(
                        {"type": "pong", "timestamp": datetime.utcnow().isoformat()}
                    )

                elif message.get("type") == "input_response":
                    # Handle input response from client
                    response = message.get("response")
                    if response is not None:
                        await ws_manager.handle_input_response(run_id, response)
                    else:
                        logger.warning(
                            f"Invalid input response format for run {run_id}"
                        )

            except json.JSONDecodeError:
                logger.warning(f"Invalid JSON received: {raw_message}")
                await websocket.send_json(
                    {
                        "type": "error",
                        "error": "Invalid message format",
                        "timestamp": datetime.utcnow().isoformat(),
                    }
                )

    except WebSocketDisconnect:
        logger.info(f"WebSocket disconnected for run {run_id}")
    except Exception as e:
        logger.error(f"WebSocket error: {str(e)}")
    finally:
        await ws_manager.disconnect(run_id)



from langchain_core.documents import Document
from langchain_core.runnables import Runnable, RunnableConfig

from mtmai.agents.ctx import mtmai_context

from mtmai.core.logging import get_logger
from mtmai.models.graph_config import ResearchState
logger = get_logger()


class IndexReferencesNode:
    def __init__(self, runnable: Runnable):
        self.runnable = runnable

    def node_name(self):
        return "index_references"

    async def __call__(self, state: ResearchState, config: RunnableConfig):
        logger.info(f"进入 index_references node: {state}")
        vs = mtmai_context .vectorstore

        all_docs = []
        for interview_state in state["interview_results"]:
            reference_docs = [
                Document(page_content=v, metadata={"source": k})
                for k, v in interview_state["references"].items()
            ]
            all_docs.extend(reference_docs)
        # await vectorstore.aadd_documents(all_docs)
        await vs.aadd_documents(all_docs)
        return state



import torch

# from transformers import DistilBertTokenizer, DistilBertForSequenceClassification
from transformers import pipeline


def tran():
    # 定义分类函数
    def classify_text(input_text, candidate_labels):
        # test_input=""I have a problem with my iphone that needs to be resolved asap!!","
        classifyResult = pipeModel1(
            input_text,
            candidate_labels=candidate_labels,
        )
        # 挑选最匹配的一个
        max_score_index = classifyResult["scores"].index(max(classifyResult["scores"]))
        max_label = classifyResult["labels"][max_score_index]
        # ret = classifyResult[0]
        return max_label

    # version = pkg_resources.get_distribution("mtmtrain").version
    # print("版本号2222： ", version)

    # version = version("mtmtrain")
    # print("版本号：：", version)

    # 固定随机数，可以确保相同环境下运算的结果相同。
    seed_value = 42
    torch.manual_seed(seed_value)

    # 使用 GPU 如果可用
    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

    # 加载预训练的 DistilBERT 分类模型
    model_name = "distilbert-base-uncased"

    # 假设 category_labels 是你的类别标签列表，例如 ["News", "Technology", "Food"]
    # category_labels = ["News", "Technology", "Food"]
    # num_labels = len(category_labels)

    # tokenizer = DistilBertTokenizer.from_pretrained(model_name)
    # tokenizer = AutoTokenizer.from_pretrained(model_name)
    # model = DistilBertForSequenceClassification.from_pretrained(model_name, num_labels=num_labels)
    # model = DistilBertForSequenceClassification.from_pretrained(model_name, problem_type="multi_label_classification")

    # model.to(device)

    pipeModel1 = pipeline(model="facebook/bart-large-mnli")



import json
import uuid

import orjson
from langchain_core.messages import AIMessage, HumanMessage, ToolMessage
from langchain_core.output_parsers import PydanticOutputParser
from langchain_core.prompts import ChatPromptTemplate, MessagesPlaceholder
from langgraph.graph.message import AnyMessage
from pydantic import BaseModel, Field

from mtmai.agents.ctx import mtmai_context
from mtmai.agents.tools.web_search import search_engine
from mtmai.core.logging import get_logger
from mtmai.llm.llm import get_fast_llm
from mtmai.models.graph_config import Editor, InterviewState

logger = get_logger()

max_str_len: int = 15000


def swap_roles(state: InterviewState, name: str):
    converted = []
    for message in state.messages:
        if isinstance(message, AIMessage) and message.name != name:
            message = HumanMessage(**message.model_dump(exclude={"type"}))
        converted.append(message)
    return InterviewState(
        messages=converted,
        references=state.references,
        editor=state.editor,
    )


def tag_with_name(ai_message: AIMessage, name: str):
    ai_message.name = name
    return ai_message


class QaNodeRequest(BaseModel):
    topic: str
    editor: Editor
    max_turns: int = 5


class QaNodeResult(BaseModel):
    editor: Editor
    messages: list[AIMessage]

    def format_conversation(self):
        messages = self.messages
        convo = "\n".join(f"{m.name}: {m.content}" for m in messages)
        return f"Conversation with {self.editor.name}\n\n" + convo


async def node_qa(req: QaNodeRequest) -> QaNodeResult:
    """
    以问答的方式发起多轮对话。
    一般应用场景是:
        1. 给定领域专家名称，提问者根据给定目的和主题发起多轮提问
    """
    topic = req.topic
    editor = req.editor

    # TODO: 提示词，需要完善
    init_messages = [
        AIMessage(
            content=f"So you said you were writing an article on {topic}?",
            name="Subject_Matter_Expert",
        )
    ]
    state = InterviewState(topic=topic, editor=editor, messages=init_messages)
    # ctx = get_mtmai_ctx()

    gen_qn_prompt = ChatPromptTemplate.from_messages(
        [
            (
                "system",
                """You are an experienced Wikipedia writer and want to edit a specific page. \
    Besides your identity as a Wikipedia writer, you have a specific focus when researching the topic. \
    Now, you are chatting with an expert to get information. Ask good questions to get more useful information.

    When you have no more questions to ask, say "Thank you so much for your help!" to end the conversation.\
    Please only ask one question at a time and don't ask what you have asked before.\
    Your questions should be related to the topic you want to write.
    Be comprehensive and curious, gaining as much unique insight from the expert as possible.\

    Stay true to your specific perspective:

    {persona}""",
            ),
            MessagesPlaceholder(variable_name="messages", optional=True),
        ]
    )

    llm = get_fast_llm()

    max_turns = req.max_turns
    for i in range(max_turns):
        # Step 1: Swap roles
        swapped_state = swap_roles(state, editor.name)

        # Step 2: Generate prompt with persona
        prompt = gen_qn_prompt.format(
            persona=editor.persona, messages=swapped_state.messages
        )

        # Step 3: Call the language model
        llm_response = await llm.ainvoke(prompt)

        # Step 4: Tag the response with the editor's name
        result = tag_with_name(llm_response, editor.name)

        state.messages.append(result)

        # 专家回答开始 ==============================================================================
        name = "Subject_Matter_Expert"
        swapped_state = swap_roles(state, name)  # Convert all other AI messages
        queries = await _gen_queries(swapped_state.messages)
        if not queries:
            logger.warning("生成搜索查询失败")
            return {"error": " gen_answer 没有生成查询"}
        # queries_input = parsed_queries.queries
        query_results = await search_engine.abatch(queries.queries)
        successful_results = [
            res for res in query_results if not isinstance(res, Exception)
        ]
        all_query_results = {
            res["url"]: res["content"]
            for results in successful_results
            for res in results
        }
        # We could be more precise about handling max token length if we wanted to here
        # Construct AI message with tool call
        tool_id = str(uuid.uuid4())[:10]
        tool_call = {
            "id": tool_id,
            "type": "function",
            "function": {
                "name": "search",
                "arguments": json.dumps({"queries": queries.queries}),
            },
        }
        ai_message = AIMessage(
            content="Searching for information...",
            additional_kwargs={"tool_calls": [tool_call]},
        )

        # Construct tool message with search results
        dumped = json.dumps(all_query_results)[:max_str_len]
        tool_message = ToolMessage(tool_call_id=tool_id, content=dumped)

        # Add both messages to the state
        state.messages.extend([ai_message, tool_message])

        answer: AnswerWithCitations = await _gen_answer(state.messages)
        cited_urls = set(answer.cited_urls)
        if cited_urls:
            # Save the retrieved information to a the shared state for future reference
            cited_references = {
                k: v for k, v in all_query_results.items() if k in cited_urls
            }
        else:
            logger.warning("没有生成引用")
            cited_references = {}
        formatted_message = AIMessage(name=name, content=answer.answer)

        data2 = {"messages": [formatted_message], "references": cited_references}

    return QaNodeResult(editor=state.editor, messages=init_messages)


class Queries(BaseModel):
    queries: list[str] = Field(
        description="Comprehensive list of search engine queries to answer the user's questions.",
    )


async def _gen_queries(messages: list[AnyMessage]):
    parser = PydanticOutputParser(pydantic_object=Queries)
    gen_queries_prompt = ChatPromptTemplate.from_messages(
        [
            (
                "system",
                "You are a helpful research assistant. Query the search engine to answer the user's questions."
                "\n\nIMPORTANT: Your response must be in valid JSON format. Follow these guidelines:"
                "\n- Use double quotes for all strings"
                "\n- Ensure all keys and values are properly enclosed"
                "\n- Do not include any text outside of the JSON object"
                "\n- Strictly adhere to the following JSON schema:"
                "\n{format_instructions}"
                "\n\nDouble-check your output to ensure it is valid JSON before submitting.",
            ),
            MessagesPlaceholder(variable_name="messages", optional=False),
        ]
    ).partial(
        format_instructions=parser.get_format_instructions(),
        messages=messages,
    )
    ai_response = await mtmai_context.call_model_chat(gen_queries_prompt, {})
    return Queries.model_validate(
        orjson.loads(mtmai_context.repair_json(ai_response.content))
    )


class AnswerWithCitations(BaseModel):
    answer: str = Field(
        description="Comprehensive answer to the user's question with citations.",
    )
    cited_urls: list[str] = Field(
        description="List of urls cited in the answer.",
    )

    @property
    def as_str(self) -> str:
        return f"{self.answer}\n\nCitations:\n\n" + "\n".join(
            f"[{i+1}]: {url}" for i, url in enumerate(self.cited_urls)
        )


async def _gen_answer(messages: list[AnyMessage]):
    parser = PydanticOutputParser(pydantic_object=AnswerWithCitations)
    gen_answer_prompt = ChatPromptTemplate.from_messages(
        [
            (
                "system",
                "You are an expert who can use information effectively. You are chatting with a Wikipedia writer who wants"
                "to write a Wikipedia page on the topic you know. You have gathered the related information and will now use the information to form a response."
                "Make your response as informative as possible and make sure every sentence is supported by the gathered information."
                "Each response must be backed up by a citation from a reliable source, formatted as a footnote, reproducing the URLS after your response."
                "\n\nIMPORTANT: Your response must be in valid JSON format. Follow these guidelines:"
                "\n- Use double quotes for all strings"
                "\n- Ensure all keys and values are properly enclosed"
                "\n- Do not include any text outside of the JSON object"
                "\n- Strictly adhere to the following JSON schema:"
                "\n{format_instructions}"
                "\n\nDouble-check your output to ensure it is valid JSON before submitting.",
            ),
            MessagesPlaceholder(variable_name="messages", optional=False),
        ]
    ).partial(
        format_instructions=parser.get_format_instructions(),
    )

    ai_response = await mtmai_context.call_model_chat(
        gen_answer_prompt,
        {
            "messages": messages,
        },
    )
    return AnswerWithCitations.model_validate(
        orjson.loads(mtmai_context.repair_json(ai_response.content))
    )



import uuid
from datetime import datetime
from enum import Enum
from typing import TYPE_CHECKING

from langchain_core.runnables import RunnableConfig
from pydantic import BaseModel
from sqlmodel import JSON, Column, Field, Relationship, SQLModel

from mtmai.mtlibs import mtutils

if TYPE_CHECKING:
    import mtmai


class AgentTaskBase(SQLModel):
    title: str | None = Field(default="")


class AgentTask(AgentTaskBase, table=True):
    """对应 langgraph 一个工作流的运行"""

    id: uuid.UUID = Field(default_factory=uuid.uuid4, primary_key=True)
    owner_id: uuid.UUID | None = Field(
        foreign_key="user.id", index=True, nullable=False, ondelete="CASCADE"
    )

    # id: str = Field(default_factory=mtutils.gen_orm_id_key, primary_key=True)
    # user_id: str = Field(default=None, foreign_key="user.id")
    created_at: datetime = Field(default_factory=datetime.now, nullable=False)
    updated_at: datetime = Field(default_factory=datetime.now, nullable=False)
    output: str | None = Field(default="")
    config: RunnableConfig = Field(sa_column=Column(JSON))


class AgentMeta(BaseModel):
    id: str
    name: str
    base_url: str
    chat_url: str | None = None
    can_chat: bool = (False,)
    agent_type: str | None = None
    graph_image: str | None = None
    label: str | None = None
    description: str | None = None


class CopilotScreen(SQLModel):
    id: str
    label: str | None = None
    Icon: str | None = None


# ui 状态相关开始
class ChatBotUiStateBase(SQLModel):
    class Config:
        arbitrary_types_allowed = True

    agent: str | None = None
    layout: str | None = None
    theme: str | None = None
    # config: dict = Field(default_factory=dict, sa_column=Column(JSON))
    isOpen: bool = Field(default=False)
    fabDisplayText: str | None = Field(default=None)
    fabDisplayIcon: str | None = Field(default=None)
    fabDisplayColor: str | None = Field(default=None)
    fabDisplayAction: str | None = Field(default=None)
    isOpenDataView: bool = Field(default=False)
    activateViewName: str | None = None
    activateChatProfileId: str | None = None
    screens: list[CopilotScreen] = Field(default_factory=list)  # 过时，


class ChatBotUiStateResponse(ChatBotUiStateBase):
    pass


class ChatBotUiState(ChatBotUiStateBase):
    id: str = Field(default_factory=mtutils.gen_orm_id_key, primary_key=True)
    user_id: str = Field(default=None, foreign_key="user.id")
    user: "mtmai.models.models.User" = Relationship(back_populates="uimessages")


class ArtifaceBase(SQLModel):
    artiface_type: str
    title: str | None = None
    description: str | None = None
    props: dict = Field(default_factory=dict, sa_column=Column(JSON))


class AgentViewType(str, Enum):
    SIDEBAR = "sidebar"  # IDE 右侧聊天样式
    POPUP = "popup"  # 页面中央弹出式


class AgentBootstrap(SQLModel):
    """前端获取 agent 的配置
    前端全局agent 加载器会在所有页面加载时返回配置初始化agent UI
    """

    # 前端agent 视图类型
    view_type: AgentViewType | None = Field(default=AgentViewType.SIDEBAR)
    # 是否显示浮动按钮
    is_show_fab: bool = Field(default=True)

    # 其他配置以后再补充


class ChatbotBase(SQLModel):
    title: str | None = Field(default="")
    description: str | None = Field(default="")
    config: dict = Field(default_factory=dict, sa_column=Column(JSON))


class Chatbot(ChatbotBase, table=True):
    id: str = Field(default_factory=mtutils.gen_orm_id_key, primary_key=True)
    created_at: datetime = Field(default_factory=datetime.now, nullable=False)
    updated_at: datetime = Field(default_factory=datetime.now, nullable=False)
    # messages: list["UiMessage"] = Relationship(back_populates="chatbot")


class UiMessageBase(SQLModel):
    class Config:
        # Needed for Column(JSON)
        arbitrary_types_allowed = True

    role: str | None = Field(default=None, max_length=64)
    content: str | None = Field(default=None)
    component: str | None = Field(default=None, max_length=64)  # 可能过时了。
    props: dict | None = Field(default=None, sa_column=Column(JSON))  # 可能过时了。
    artifacts: list[dict] | None = Field(default=None, sa_column=Column(JSON))


class UiMessage(UiMessageBase, table=True):
    """前端 聊天机器人的消息列表组件"""

    id: str = Field(default_factory=mtutils.gen_orm_id_key, primary_key=True)
    chatbot_id: str = Field(default=None, foreign_key="chatbot.id")
    created_at: datetime = Field(default_factory=datetime.now, nullable=False)
    updated_at: datetime = Field(default_factory=datetime.now, nullable=False)



import json
import typing

import pydantic.json
import structlog

from mtmai.forge.sdk.artifact.models import Artifact, ArtifactType
from mtmai.forge.sdk.db.enums import OrganizationAuthTokenType
from mtmai.forge.sdk.db.models import (
    ArtifactModel,
    AWSSecretParameterModel,
    BitwardenLoginCredentialParameterModel,
    BitwardenSensitiveInformationParameterModel,
    OrganizationAuthTokenModel,
    OrganizationModel,
    OutputParameterModel,
    StepModel,
    TaskModel,
    WorkflowModel,
    WorkflowParameterModel,
    WorkflowRunModel,
    WorkflowRunOutputParameterModel,
    WorkflowRunParameterModel,
)
from mtmai.forge.sdk.models import Organization, OrganizationAuthToken, Step, StepStatus
from mtmai.forge.sdk.schemas.tasks import ProxyLocation, Task, TaskStatus
from mtmai.forge.sdk.workflow.models.parameter import (
    AWSSecretParameter,
    BitwardenLoginCredentialParameter,
    BitwardenSensitiveInformationParameter,
    OutputParameter,
    WorkflowParameter,
    WorkflowParameterType,
)
from mtmai.forge.sdk.workflow.models.workflow import (
    Workflow,
    WorkflowDefinition,
    WorkflowRun,
    WorkflowRunOutputParameter,
    WorkflowRunParameter,
    WorkflowRunStatus,
)
from mtmai.models.task import MtTask

LOG = structlog.get_logger()


@typing.no_type_check
def _custom_json_serializer(*args, **kwargs) -> str:
    """
    Encodes json in the same way that pydantic does.
    """
    return json.dumps(*args, default=pydantic.json.pydantic_encoder, **kwargs)


def convert_to_mttask(task_obj: MtTask, debug_enabled: bool = False) -> Task:
    if debug_enabled:
        LOG.debug("Converting TaskModel to Task", task_id=task_obj.task_id)
    task = Task(
        task_id=task_obj.task_id,
        status=TaskStatus(task_obj.status),
        created_at=task_obj.created_at,
        modified_at=task_obj.modified_at,
        title=task_obj.title,
        url=task_obj.url,
        webhook_callback_url=task_obj.webhook_callback_url,
        totp_verification_url=task_obj.totp_verification_url,
        totp_identifier=task_obj.totp_identifier,
        navigation_goal=task_obj.navigation_goal,
        data_extraction_goal=task_obj.data_extraction_goal,
        navigation_payload=task_obj.navigation_payload,
        extracted_information=task_obj.extracted_information,
        failure_reason=task_obj.failure_reason,
        organization_id=task_obj.organization_id,
        proxy_location=(
            ProxyLocation(task_obj.proxy_location) if task_obj.proxy_location else None
        ),
        extracted_information_schema=task_obj.extracted_information_schema,
        workflow_run_id=task_obj.workflow_run_id,
        order=task_obj.order,
        retry=task_obj.retry,
        max_steps_per_run=task_obj.max_steps_per_run,
        error_code_mapping=task_obj.error_code_mapping,
        errors=task_obj.errors,
    )
    return task



from phi.agent import Agent
from phi.embedder.together import TogetherEmbedder
from phi.knowledge.pdf import PDFUrlKnowledgeBase
from phi.tools.crawl4ai_tools import Crawl4aiTools
from phi.vectordb.pgvector import PgVector, SearchType

from mtmai.agents.phiagents.config import model
from mtmai.core.config import settings
from mtmai.core.logging import get_logger

logger = get_logger()


url_pdf_knowledge_base = PDFUrlKnowledgeBase(
    urls=["https://phi-public.s3.amazonaws.com/recipes/ThaiRecipes.pdf"],
    vector_db=PgVector(
        table_name="recipes_demo234",
        db_url=settings.MTMAI_DATABASE_URL,
        search_type=SearchType.hybrid,
        embedder=TogetherEmbedder(
            api_key="10747773f9883cf150558aca1b0dda81af4237916b03d207b8ce645edb40a546",
            # dimensions=1536,
        ),
    ),
)
# url_pdf_knowledge_base.load(upsert=True)


def get_current_user_site_id() -> str:
    """Get site id for current user.

    Args:
        no args required

    Returns:
        str:    .
    """
    return "fake-site-id-123456"


workbrench_agent = Agent(
    name="Web Container Agent",
    role="work branch assistant",
    # knowledge=url_pdf_knowledge_base,
    model=model,
    tools=[Crawl4aiTools(max_length=10000)],
    markdown=True,
    add_datetime_to_instructions=True,
    show_tool_calls=True,
    description="你是Bolt,是专业有用的助手，会进行工作流程的思考,协助用户完成**博客网站文章自动生成和发布**任务,能使用知识库回答用户问题",
    instructions=[
        """
<background_story>
  你运行在前端浏览器中，主聊天窗口就是你，用户可以直接跟你对话，通过约定格式生成多个约定格式的指令协助用户完成site项目的操作和配置
  聊天窗口在左侧，可折叠和展开
  工作区在右侧，可以通过指令进行展开和折叠，是用户可以自由操作的部分，就像一个常见的多模块后台，可以进行数据查询和编辑
  前端会根据情况将 params 参数自动填充到 <params> 标签中，你只需要在需要的时候使用即可。
  params 参数表示当前页面的一些参数，常见的有
      siteId: 当前操作的站点Id
      userId: 当前用户Id
</background_story>
{workbench_prompt}
<system_constraints>
  * IMPORTANT: 文章生成任务必须有 siteId，因此如果缺少 siteId 参数，你应该使用 <askhuman> 询问用户，并且等待用户回复。
  * IMPORTANT: 文章是属于复杂结构的文章，所以你不要直接输出文章而是调用后端的工作流。
  * 根据siteId, 在必要时，可以调用相关的 工具获取 站点信息，或者调用相关的工作流 进行操作。

    工作流的运行步骤和结果会实时反应在UI当中，并且会在后续聊天消息中附带工作流的状态和结果。
    一般流程:
      - 根据用户的聊天消息识别用户的意图
      - 根据意图调用后端的工作流
      - 如果工作流启动成功，用户可以在UI上看见状态，并可以选择取消，暂停，继续等操作。
      - 工作流结束后会在UI上显示最终结果，并且在聊天窗口的右侧可以详细查看和编辑工作流输出构件。
      - 用户编辑构件如果存在疑问、对结果不满意、可以通过聊天窗口反馈给你，并且反馈的消息会附带工作流基本数据或者完整数据，你需要根据用户的反馈做出解答，或者调用函数、工具等进一步操作。
      - 用户跟你聊天你应该针对上下文主题认真全面思考做出答复，告诉用户不要闲聊，如果用户的要求明显超出了工作流的主题意图，应该建议他重启新的任务。
      - 工作流的输出构件，是源码，例如生成了一篇文章就是markdown格式的源码，UI会进行渲染，但是你应该只关注源码，后续的操作你可以通过常见的开发流程使用 diff patch 的方式对源码进行修改。
      - 一个对话过程本质就是围绕构件进行，并且同一时间只可能有一个构件。用户如果重做的要求，就是调用新的工作流，工作流输出的构件会自动覆盖旧的。
</system_constraints>
"""
    ],
)



"""
项目参考:
    https://github.com/SeleniumHQ/docker-selenium
"""

import atexit
import os
import subprocess
import time
from pathlib import Path
import os

from mtmai.core.config import settings
from mtmai.core.logging import get_logger
from mtmlib.httpUtils import download_file
from mtmlib.mtutils import is_in_gitpod

logger = get_logger()

selenium_server_jar_bin = str(
    Path(settings.storage_dir).joinpath("seleniumselenium-server.jar")
)
selenium_version = "4.24.0"
selenium_server_port = 4444


async def install_selenium_server():
    if not Path(selenium_server_jar_bin).exists():
        jar_url = f"https://github.com/SeleniumHQ/selenium/releases/download/selenium-{selenium_version}/selenium-server-{selenium_version}.jar"
        await download_file(jar_url, selenium_server_jar_bin)


async def install_selenium():
    try:
        # Install Selenium and its dependencies
        subprocess.run(["pip", "install", "selenium", "webdriver_manager"], check=True)

        # Create necessary directories
        directories = [
            "/opt/selenium",
            "/opt/selenium/assets",
            "/opt/selenium/secrets",
            "/var/run/supervisor",
            "/var/log/supervisor",
            os.environ.get("SEL_DOWNLOAD_DIR", "/tmp/selenium"),
            f"{os.environ['HOME']}/.mozilla",
            f"{os.environ['HOME']}/.vnc",
            f"{os.environ['HOME']}/.pki/nssdb",
        ]
        for directory in directories:
            Path(directory).mkdir(parents=True, exist_ok=True)

        # Initialize NSSDB with an empty password
        subprocess.run(
            [
                "certutil",
                "-d",
                f"sql:{os.environ['HOME']}/.pki/nssdb",
                "-N",
                "--empty-password",
            ],
            check=True,
        )

        # Create config file
        Path("/opt/selenium/config.toml").touch()

        # Set permissions
        sel_user = os.environ.get("SEL_USER", "seluser")
        sel_group = os.environ.get("SEL_GROUP", "seluser")
        subprocess.run(
            [
                "chown",
                "-R",
                f"{sel_user}:{sel_group}",
                "/opt/selenium",
                "/var/run/supervisor",
                "/var/log/supervisor",
                "/etc/passwd",
                os.environ["HOME"],
            ],
            check=True,
        )
        subprocess.run(
            [
                "chmod",
                "-R",
                "775",
                "/opt/selenium",
                "/var/run/supervisor",
                "/var/log/supervisor",
                "/etc/passwd",
                os.environ["HOME"],
            ],
            check=True,
        )

        # Download Selenium server
        authors = os.environ.get("AUTHORS", "SeleniumHQ")
        release = os.environ.get("RELEASE", "4.10.0")
        version = os.environ.get("VERSION", "4.10.0")
        await download_file(
            f"https://github.com/{authors}/selenium/releases/download/{release}/selenium-server-{version}.jar",
            "/opt/selenium/selenium-server.jar",
        )

        # Set group permissions
        subprocess.run(
            [
                "chgrp",
                "-R",
                "0",
                "/opt/selenium",
                os.environ["HOME"],
                "/opt/selenium/assets",
                "/var/run/supervisor",
                "/var/log/supervisor",
            ],
            check=True,
        )
        subprocess.run(
            [
                "chmod",
                "-R",
                "g=u",
                "/opt/selenium",
                os.environ["HOME"],
                "/opt/selenium/assets",
                "/var/run/supervisor",
                "/var/log/supervisor",
            ],
            check=True,
        )

        # Set ACLs
        for path in [
            "/opt",
            "/opt/selenium",
            os.environ["HOME"],
            "/opt/selenium/assets",
            "/var/run/supervisor",
            "/var/log/supervisor",
        ]:
            subprocess.run(["setfacl", "-Rm", f"u:{sel_user}:rwx", path], check=True)
            subprocess.run(["setfacl", "-Rm", f"g:{sel_group}:rwx", path], check=True)

        logger.info("Selenium and necessary components installed successfully.")
    except subprocess.CalledProcessError as e:
        logger.error(f"An error occurred during installation: {e}")
    except Exception as e:
        logger.error(f"An unexpected error occurred: {e}")


def is_desktop_environment():
    # Check for common desktop environment variables
    desktop_vars = [
        "XDG_CURRENT_DESKTOP",
        "GNOME_DESKTOP_SESSION_ID",
        "KDE_FULL_SESSION",
        "DESKTOP_SESSION",
    ]
    return any(var in os.environ for var in desktop_vars)


def is_x11_running():
    try:
        # Check if X11 is running
        subprocess.run(
            ["xset", "q"],
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
            check=True,
        )
        return True
    except (subprocess.CalledProcessError, FileNotFoundError):
        return False


def is_in_display():
    """是否处于图形界面环境中"""
    return is_desktop_environment() or is_x11_running()


async def start_selenium_server():
    logger.info("Starting Selenium server ...")
    logger.info("settings.storage_dir: %s", settings.storage_dir)
    logger.info("selenium_server_jar_bin: %s", selenium_server_jar_bin)
    logger.info("selenium_version: %s", selenium_version)
    logger.info("selenium_server_port: %s", selenium_server_port)
    logger.info("is_in_display: %s", is_in_display())

    await install_selenium_server()
    if not is_in_display():
        logger.info("当前非图形界面，通过 $DISPLAY 环境变量指定图形界面")
        if is_in_gitpod():
            """特定环境下自动配置"""
            os.environ["DISPLAY"] = ":1"

    extra_envs = {
        # "DISPLAY": ":1",
        # "JAVA_OPTS": "-Dwebdriver.chrome.whitelistedIps=",
        # "SE_NODE_MAX_SESSIONS": "8",
        # "SE_NODE_OVERRIDE_MAX_SESSIONS": "true",
        # "SE_NODE_SESSION_TIMEOUT": "300",
        # "SE_OPTS": "--log-level INFO"
    }


    selenium_command = [
        "java",
        "-jar",
        selenium_server_jar_bin,
        "standalone",
        "--selenium-manager",
        "true",
    ]
    env = os.environ.copy()
    env.update(extra_envs)
    selenium_process = subprocess.Popen(
        selenium_command,
        # env=env
    )

    atexit.register(lambda: selenium_process.terminate())

    # time.sleep(7)
    # logger.info(
    #     f"🚀 Selenium server started successfully!, port : {selenium_server_port}"
    # )

    # Monitor the Selenium server process
    while True:
        if selenium_process.poll() is not None:
            logger.error("Selenium server process has terminated unexpectedly.")
            break
        time.sleep(1)


def test_selenium():
    from selenium.webdriver.common.by import By
    from selenium.webdriver.support import expected_conditions as EC
    from selenium.webdriver.support.ui import WebDriverWait

    driver = start_selenium_server()
    if driver:
        try:
            driver.get("https://www.bing.com")
            WebDriverWait(driver, 10).until(
                EC.presence_of_element_located((By.TAG_NAME, "body"))
            )
            logger.info("Successfully accessed Bing.com")
            return True
        except Exception as e:
            logger.error(f"Failed to access Bing.com: {e}")
            return False
        finally:
            driver.quit()
    else:
        logger.error("Failed to start Selenium")
        return False



import uuid
from datetime import datetime
from enum import Enum

from pgvector.sqlalchemy import Vector
from pydantic import EmailStr
from sqlalchemy import PrimaryKeyConstraint, UniqueConstraint
from sqlmodel import JSON, Column, Field, Relationship, SQLModel


# Shared properties
class UserBase(SQLModel):
    email: EmailStr = Field(unique=True, index=True, max_length=255)
    is_active: bool = True
    is_superuser: bool = False
    full_name: str | None = Field(default=None, max_length=255)
    username: str | None = Field(default=None, max_length=255)
    identifier: str | None = Field(default=None, max_length=255)
    meta: dict | None = Field(default={}, sa_column=Column(JSON))
    organization_id: str | None = Field(default=None, max_length=255)


# Properties to receive via API on creation
class UserCreate(UserBase):
    password: str = Field(min_length=6, max_length=40)


class UserRegister(SQLModel):
    email: EmailStr = Field(max_length=255)
    password: str = Field(min_length=6, max_length=40)
    full_name: str | None = Field(default=None, max_length=255)


# Properties to receive via API on update, all are optional
class UserUpdate(UserBase):
    email: EmailStr | None = Field(default=None, max_length=255)  # type: ignore
    password: str | None = Field(default=None, min_length=6, max_length=40)


class UserUpdateMe(SQLModel):
    full_name: str | None = Field(default=None, max_length=255)
    email: EmailStr | None = Field(default=None, max_length=255)


class UpdatePassword(SQLModel):
    current_password: str = Field(min_length=8, max_length=40)
    new_password: str = Field(min_length=6, max_length=40)


class User(UserBase, table=True):
    id: uuid.UUID = Field(default_factory=uuid.uuid4, primary_key=True)
    # id: str = Field(default_factory=mtutils.gen_orm_id_key, primary_key=True)
    hashed_password: str
    created_at: datetime = Field(default_factory=datetime.now, nullable=False)
    updated_at: datetime = Field(default_factory=datetime.now, nullable=False)
    items: list["Item"] = Relationship(back_populates="owner", cascade_delete=True)
    # documents: list["Document"] = Relationship(
    #     back_populates="owner", cascade_delete=True
    # )
    # doccolls: list["DocColl"] = Relationship(
    #     back_populates="owner", cascade_delete=True
    # )
    account: "Account" = Relationship(back_populates="owner", cascade_delete=True)

    # chats: "mtmai.models.chat.ChatInput" = Relationship(
    #     back_populates="user", cascade_delete=True
    # )
    # agenttasks: "mtmai.models.agent.AgentTask" = Relationship(
    #     back_populates="user", cascade_delete=True
    # )
    # uimessages: "mtmai.models.agent.UiMessage" = Relationship(
    #     back_populates="user", cascade_delete=True
    # )


# Properties to return via API, id is always required
class UserPublic(UserBase):
    id: uuid.UUID


class UsersPublic(SQLModel):
    data: list[UserPublic]
    count: int


# account ##################################################################################
class AccountBase(SQLModel):
    provider: str
    token: str


class Account(AccountBase, table=True):
    # id: str = Field(default_factory=mtutils.gen_orm_id_key, primary_key=True)
    id: uuid.UUID = Field(default_factory=uuid.uuid4, primary_key=True)
    # owner_id: str = Field(foreign_key="user.id", nullable=False, ondelete="CASCADE")
    owner_id: uuid.UUID | None = Field(
        foreign_key="user.id", index=True, nullable=False, ondelete="CASCADE"
    )

    owner: User | None = Relationship(back_populates="account")


# items ####################################################################################
# Shared properties
class ItemBase(SQLModel):
    title: str = Field(min_length=1, max_length=255)
    description: str | None = Field(default=None, max_length=255)


# Properties to receive on item creation
class ItemCreate(ItemBase):
    pass


# Properties to receive on item update
class ItemUpdate(ItemBase):
    title: str | None = Field(default=None, min_length=1, max_length=255)  # type: ignore


# Database model, database table inferred from class name
class Item(ItemBase, table=True):
    # id: uuid.UUID = Field(default_factory=uuid.uuid4, primary_key=True)
    # id: str = Field(default_factory=mtutils.gen_orm_id_key, primary_key=True)
    id: uuid.UUID = Field(default_factory=uuid.uuid4, primary_key=True)
    title: str = Field(max_length=255)
    # owner_id: str = Field(foreign_key="user.id", nullable=False, ondelete="CASCADE")
    owner_id: uuid.UUID | None = Field(
        foreign_key="user.id", index=True, nullable=False, ondelete="CASCADE"
    )
    owner: User | None = Relationship(back_populates="items")


# Properties to return via API, id is always required
class ItemPublic(ItemBase):
    id: uuid.UUID
    owner_id: uuid.UUID


class ItemsPublic(SQLModel):
    data: list[ItemPublic]
    count: int


# Generic message
class Message(SQLModel):
    message: str


# JSON payload containing access token
class Token(SQLModel):
    access_token: str
    token_type: str = "bearer"


# Contents of JWT token
class TokenPayload(SQLModel):
    sub: str | None = None


class NewPassword(SQLModel):
    token: str
    new_password: str = Field(min_length=8, max_length=40)


# ---------------------------------------------------------------------------------------------------------------------


class StatusEnum(str, Enum):
    """工作流状态"""

    NEW = "new"  # New state
    PENDING = "pending"  # Waiting for execution
    RUNNING = "running"
    WAITING_FOR_HUMAN = "waiting_for_human"  # Awaiting human intervention
    CONTINUE_AFTER_CONFIRMATION = (
        "continue_after_confirmation"  # Continue after confirmation
    )
    # START = "start"
    END = "end"
    PAUSE = "pause"


# 知识库相关
class DocumentBase(SQLModel):
    class Config:
        arbitrary_types_allowed = True

    title: str | None = Field(default=None, max_length=255)

    collection: str = Field(default="default", index=True)
    meta: dict | None = Field(default={}, sa_column=Column(JSON))
    content: str | None = Field(default=None, max_length=8196)
    content_type: str | None = Field(default=None, max_length=255)


class Document(DocumentBase, table=True):
    """
    通用的基于 postgres + pgvector 的 rag 文档
    注意: 需要提前运行: CREATE EXTENSION IF NOT EXISTS vector
    参考: https://github.com/pgvector/pgvector-python/tree/master
    """

    id: uuid.UUID = Field(default_factory=uuid.uuid4, primary_key=True)
    embedding: list[float] = Field(sa_column=Column(Vector(1024)))
    # owner_id: str = Field(index=True, nullable=True)
    owner_id: uuid.UUID | None = Field(
        foreign_key="user.id", index=True, nullable=False, ondelete="CASCADE"
    )
    site_id: str = Field(index=True, nullable=True)
    agent_id: str = Field(index=True, nullable=True)
    created_at: datetime = Field(default_factory=datetime.now, nullable=False)
    updated_at: datetime = Field(default_factory=datetime.now, nullable=False)
    is_deleted: bool = Field(default=False, index=True)


class SysItem(SQLModel, table=True):
    type: str = Field(index=True)
    key: str = Field(index=True)
    value: str
    description: str | None = Field(default=None, max_length=255)

    __table_args__ = (
        PrimaryKeyConstraint("type", "key"),
        UniqueConstraint("type", "key", name="sys_item_type_key_uc"),
    )

    class Config:
        table_name = "sys_item"
